from unittest import TestCase

from datetime import datetime

import pytz
from freezegun import freeze_time
from hypothesis import given
from hypothesis.extra.pytz import timezones
from hypothesis.strategies import datetimes, decimals, floats, integers, one_of, sampled_from

from oasislmf.utils.values import get_timestamp, get_utctimestamp, to_string, NULL_VALUES, to_int, to_float


class GetTimestamp(TestCase):
    def test_time_is_not_supplied___timestamp_of_now_is_returned(self):
        _now = datetime.now()

        with freeze_time(_now):
            res = get_timestamp()

            self.assertEqual(_now.strftime('%Y%m%d%H%M%S'), res)

    @given(datetimes(min_value=datetime(1990, 1, 1)))
    def test_time_is_supplied___timestamp_of_supplied_time_is_returned(self, dt):
        res = get_timestamp(dt)

        self.assertEqual(dt.strftime('%Y%m%d%H%M%S'), res)

    @given(datetimes(min_value=datetime(1990, 1, 1)))
    def test_fmt_is_supplied___timestamp_of_supplied_time_is_returned(self, dt):
        fmt = 'foo %Y-%m-%d %H:%M:%S bar'

        res = get_timestamp(dt, fmt=fmt)

        self.assertEqual(dt.strftime(fmt), res)


class GetUtcTimestamp(TestCase):
    def test_time_is_not_supplied___timestamp_of_utcnow_is_returned(self):
        _now = datetime.utcnow()

        with freeze_time(_now):
            res = get_utctimestamp()

            self.assertEqual(_now.strftime('%Y-%b-%d %H:%M:%S'), res)

    @given(datetimes(timezones=timezones(), min_value=datetime(1990, 1, 1)))
    def test_time_is_supplied___timestamp_of_supplied_time_is_returned(self, dt):
        res = get_utctimestamp(dt)

        self.assertEqual(dt.astimezone(pytz.utc).strftime('%Y-%b-%d %H:%M:%S'), res)

    @given(datetimes(timezones=timezones(), min_value=datetime(1990, 1, 1)))
    def test_fmt_is_supplied___timestamp_of_supplied_time_is_returned(self, dt):
        fmt = 'foo %Y-%m-%d %H:%M:%S bar'

        res = get_utctimestamp(dt, fmt=fmt)

        self.assertEqual(dt.astimezone(pytz.utc).strftime(fmt), res)


class ToString(TestCase):
    def test_value_is_none___result_is_empty(self):
        res = to_string(None)

        self.assertEqual('', res)

    @given(one_of(integers(), floats(allow_infinity=False, allow_nan=False), decimals(allow_infinity=False, allow_nan=False)))
    def test_value_is_not_none___result_is_string_conversion(self, value):
        res = to_string(value)

        self.assertEqual(str(value), res)


class ToInt(TestCase):
    @given(sampled_from(NULL_VALUES))
    def test_value_is_one_of_the_none_values___result_is_none(self, value):
        res = to_int(value)

        self.assertIsNone(res)

    @given(floats(allow_infinity=False, allow_nan=False))
    def test_value_is_float___result_is_converted_to_int(self, value):
        res = to_int(value)

        self.assertEqual(int(value), res)


class ToFloat(TestCase):
    @given(sampled_from(NULL_VALUES))
    def test_value_is_one_of_the_none_values___result_is_none(self, value):
        res = to_float(value)

        self.assertIsNone(res)

    @given(decimals(allow_infinity=False, allow_nan=False))
    def test_value_is_decimal___result_is_converted_to_float(self, value):
        res = to_float(value)

        self.assertEqual(float(value), res)



from unittest import TestCase

import os
import six
from backports.tempfile import TemporaryDirectory
from mock import patch

from oasislmf.cmd import RootCmd


def get_command(target_dir=None, extras=None):
    kwargs_str = ' '.join('--{} {}'.format(k, v) for k, v in six.iteritems(extras or {}))

    return RootCmd(argv='bin clean {} {}'.format(kwargs_str, target_dir or '').split())


class CleanCmdRun(TestCase):
    @patch('oasislmf.cmd.bin.cleanup_bin_directory')
    def test_src_is_not_supplied___cwd_is_cleaned(self, clean_mock):
        cmd = get_command()

        res = cmd.run()

        self.assertEqual(0, res)
        clean_mock.assert_called_once_with(os.path.abspath('.'))

    @patch('oasislmf.cmd.bin.cleanup_bin_directory')
    def test_src_is_supplied___supplied_path_is_cleaned(self, clean_mock):
        with TemporaryDirectory() as d:
            cmd = get_command(target_dir=d)

            res = cmd.run()

            self.assertEqual(0, res)
            clean_mock.assert_called_once_with(os.path.abspath(d))



# -*- coding: utf-8 -*-

import io
import json
import os
import subprocess

from argparse import RawDescriptionHelpFormatter

from pathlib2 import Path

from ..exposures.csv_trans import Translator
from ..exposures.manager import OasisExposuresManager

from ..model_execution.bash import genbash
from ..model_execution.runner import run
from ..model_execution.bin import create_binary_files, prepare_model_run_directory, prepare_model_run_inputs

from ..utils.exceptions import OasisException
from ..utils.peril import PerilAreasIndex
from ..utils.values import get_utctimestamp

from ..keys.lookup import OasisLookupFactory

from .cleaners import as_path

from .base import OasisBaseCommand, InputValues

class GeneratePerilAreasRtreeFileIndexCmd(OasisBaseCommand):
    """
    Generates and writes an Rtree file index of peril area IDs (area peril IDs)
    and area polygon bounds from a peril areas (area peril) file.
    """
    formatter_class = RawDescriptionHelpFormatter

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(self.__class__, self).add_args(parser)

        parser.add_argument(
            '-c', '--lookup-config-file-path', default=None,
            help='Lookup config file path',
        )
        parser.add_argument(
            '-d', '--keys-data-path', default=None,
            help='Keys data path'
        )
        parser.add_argument(
            '-f', '--index-file-path', default=None,
            help='Index file path (no file extension required)',
        )

    def action(self, args):
        """
        Generates and writes an Rtree file index of peril area IDs (area peril IDs)
        and area polygon bounds from a peril areas (area peril) file.

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)
        
        lookup_config_fp = as_path(inputs.get('lookup_config_file_path', required=True, is_path=True), 'Lookup config file path', preexists=True)

        keys_data_path = as_path(inputs.get('keys_data_path', required=True, is_path=True), 'Keys config file path', preexists=True)

        index_fp = as_path(inputs.get('index_file_path', required=True, is_path=True, default=None), 'Index output file path', preexists=False)

        with io.open(lookup_config_fp, 'r', encoding='utf-8') as f:
            config = json.load(f)

        peril_config = config.get('peril')

        if not peril_config:
            raise OasisException(
                'The lookup config must contain a peril-related subdictionary with a key named '
                '`peril` defining area-peril-related model information'
            )

        areas_fp = peril_config.get('file_path')

        if not areas_fp:
            raise OasisException(
                'The lookup peril config must define the path of a peril areas '
                '(or area peril) file with the key name `file_path`'
            )

        if areas_fp.startswith('%%KEYS_DATA_PATH%%'):
            areas_fp = areas_fp.replace('%%KEYS_DATA_PATH%%', keys_data_path)

        areas_fp = as_path(areas_fp, 'areas_fp')

        src_type = str.lower(str(peril_config.get('file_type')) or '') or 'csv'

        peril_id_col = str.lower(str(peril_config.get('peril_id_col')) or '') or 'peril_id'

        coverage_config = config.get('coverage')

        if not coverage_config:
            raise OasisException(
                'The lookup config must contain a coverage-related subdictionary with a key named '
                '`coverage` defining coverage related model information'
            )

        coverage_type_col = str.lower(str(coverage_config.get('coverage_type_col')) or '') or 'coverage_type'

        peril_area_id_col = str.lower(str(peril_config.get('peril_area_id_col')) or '') or 'area_peril_id'

        area_poly_coords_cols = peril_config.get('area_poly_coords_cols')

        if not area_poly_coords_cols:
            raise OasisException(
                'The lookup peril config must define the column names of '
                'the coordinates used to define areas in the peril areas '
                '(area peril) file using the key `area_poly_coords_cols`'
            )

        non_na_cols = (
            tuple(col.lower() for col in peril_config['non_na_cols']) if peril_config.get('non_na_cols')
            else tuple(col.lower() for col in [peril_area_id_col] + area_poly_coords_cols.values())
        )

        col_dtypes = peril_config.get('col_dtypes') or {peril_area_id_col: int}

        sort_col = peril_config.get('sort_col') or peril_area_id_col

        area_poly_coords_seq_start_idx = peril_config.get('area_poly_coords_seq_start_idx') or 1

        area_reg_poly_radius = peril_config.get('area_reg_poly_radius') or 0.00166

        index_props = peril_config.get('rtree_index')
        index_props.pop('filename')

        self.logger.info(
            '\nGenerating Rtree file index {}.{{idx,dat}} from peril areas (area peril) '
            'file {}'
            .format(os.path.join(index_fp), areas_fp)
        )

        index_fp = PerilAreasIndex.create_from_peril_areas_file(
            src_fp=areas_fp,
            src_type=src_type,
            peril_id_col=peril_id_col,
            coverage_type_col=coverage_type_col,
            peril_area_id_col=peril_area_id_col,
            non_na_cols=non_na_cols,
            col_dtypes=col_dtypes,
            sort_col=sort_col,
            area_poly_coords_cols=area_poly_coords_cols,
            area_poly_coords_seq_start_idx=area_poly_coords_seq_start_idx,
            area_reg_poly_radius=area_reg_poly_radius,
            index_fp=index_fp,
            index_props=index_props
        )

        self.logger.info('\nSuccessfully generated index files {}.{{idx.dat}}'.format(index_fp))


class TransformSourceToCanonicalFileCmd(OasisBaseCommand):
    """
    Transform a supplier-specific source exposures/accounts file to a canonical
    Oasis exposures/accounts file using an XSD validation file and an XSLT
    transformation file.

    Calling syntax is::

        oasislmf model transform-source-to-canonical
            -s <source file path>
            -y <type of source file - exposures or accounts>
            -v <validation file>
            -x <transformation file>
            -o <output file path>
    """
    formatter_class = RawDescriptionHelpFormatter

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(self.__class__, self).add_args(parser)

        parser.add_argument(
            '-s', '--source-file-path', default=None,
            help='Source file path',
        )
        parser.add_argument(
            '-y', '--source-file-type', default='exposures',
            help='Type of source file - exposures or accounts',
        )
        parser.add_argument(
            '-v', '--xsd-validation-file-path', default=None,
            help='XSD validation file path',
        )
        parser.add_argument(
            '-x', '--xslt-transformation-file-path', default=None,
            help='XSLT transformation file path',
        )
        parser.add_argument(
            '-o', '--output-file-path', default=None,
            help='Output file path',
        )

    def action(self, args):
        """
        Transform a source exposures/accounts file to a canonical Oasis file.

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)
        
        source_file_path = as_path(inputs.get('source_file_path', required=True, is_path=True), 'Source file path', preexists=True)
        source_file_type = inputs.get('source_file_type', default='exposures')
        
        _sft = 'exp' if source_file_type == 'exposures' else 'acc'
        _utc = get_utctimestamp(fmt='%Y%m%d%H%M%S')
        
        xsd_validation_file_path = as_path(inputs.get('xsd_validation_file_path', required=True, is_path=True), 'XSD validation file path', preexists=True)
        xslt_transformation_file_path = as_path(inputs.get('xslt_transformation_file_path', required=True, is_path=True), 'XSLT transformation file path', preexists=True)
        output_file_path = as_path(inputs.get('output_file_path', required=False, is_path=True, default='can{}-{}.csv'.format(_sft, _utc)), 'Output file path', preexists=False)

        self.logger.info('\nGenerating a canonical {} file {} from source {} file {}'.format(_sft, output_file_path, _sft, source_file_path))

        translator = Translator(source_file_path, output_file_path, xsd_validation_file_path, xslt_transformation_file_path, append_row_nums=True)
        translator()

        self.logger.info('\nOutput file {} successfully generated'.format(output_file_path))


class TransformCanonicalToModelFileCmd(OasisBaseCommand):
    """
    Transform a canonical Oasis exposures file to a model Oasis exposures file
    using an XSD validation file and an XSLT transformation file. A model exposures
    file is a simplified version of the canonical exposures file and provides the
    input to an Oasis keys server for a model and its keys lookup class.

    Calling syntax is::

        oasislmf model transform-canonical-to-model
            -c <canonical exposures file path>
            -v <validation file>
            -x <transformation file>
            -o <output file path>
    """
    formatter_class = RawDescriptionHelpFormatter

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(self.__class__, self).add_args(parser)

        parser.add_argument(
            '-c', '--canonical-exposures-file-path', default=None,
            help='Canonical file path',
        )
        parser.add_argument(
            '-v', '--xsd-validation-file-path', default=None,
            help='XSD validation file path',
        )
        parser.add_argument(
            '-x', '--xslt-transformation-file-path', default=None,
            help='XSLT transformation file path',
        )
        parser.add_argument(
            '-o', '--output-file-path', default=None,
            help='Output file path',
        )

    def action(self, args):
        """
        Transform a source exposures/accounts file to a canonical Oasis file.

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)
        
        canonical_exposures_file_path = as_path(inputs.get('canonical_exposures_file_path', required=True, is_path=True), 'Canonical exposures file path', preexists=True)

        _utc = get_utctimestamp(fmt='%Y%m%d%H%M%S')
        
        xsd_validation_file_path = as_path(inputs.get('xsd_validation_file_path', required=True, is_path=True), 'XSD validation file path', preexists=True)
        xslt_transformation_file_path = as_path(inputs.get('xslt_transformation_file_path', required=True, is_path=True), 'XSLT transformation file path', preexists=True)
        output_file_path = as_path(inputs.get('output_file_path', required=False, is_path=True, default='modexp-{}.csv'.format(_utc)), 'Output file path', preexists=False)

        self.logger.info('\nGenerating a model exposures file {} from canonical exposures file {}'.format(output_file_path, canonical_exposures_file_path))

        translator = Translator(canonical_exposures_file_path, output_file_path, xsd_validation_file_path, xslt_transformation_file_path, append_row_nums=True)
        translator()

        self.logger.info('\nOutput file {} successfully generated'.format(output_file_path))


class GenerateKeysCmd(OasisBaseCommand):
    """
    Generate Oasis keys records (location records with area peril ID and
    vulnerability ID) for a model, and also writes them to file.

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    Run ``oasislmf config --help`` for more information.

    Keys records returned by an Oasis keys lookup service (see the PiWind
    lookup service for reference) will be Python dicts with the following
    structure
    ::

        {
            "id": <loc. ID>,
            "peril_id": <Oasis peril type ID - oasislmf/utils/peril.py>,
            "coverage": <Oasis coverage type ID - see oasislmf/utils/coverage.py>,
            "area_peril_id": <area peril ID>,
            "vulnerability_id": <vulnerability ID>,
            "message": <lookup status message>,
            "status": <lookup status code - see oasislmf/utils/status.py>
        }

    The command can generate keys records in this format, and write them to file.

    For model loss calculations however ``ktools`` requires a keys CSV file with
    the following format
    ::

        LocID,PerilID,CoverageID,AreaPerilID,VulnerabilityID
        ..
        ..

    where the headers correspond to the relevant Oasis keys record fields.
    The command can also generate and write Oasis keys files. It will also
    optionally write an Oasis keys error file, which is a file containing
    keys records for locations with unsuccessful lookups (locations with
    a failing or non-matching lookup). It has the format
    ::

        LocID,PerilID,CoverageID,Message
    """
    formatter_class = RawDescriptionHelpFormatter

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(self.__class__, self).add_args(parser)

        parser.add_argument('-k', '--keys-file-path', default=None, help='Keys file path')
        parser.add_argument('-e', '--keys-errors-file-path', default=None, help='Keys errors file path')
        parser.add_argument('-g', '--lookup-config-file-path', default=None, help='Lookup config JSON file path')
        parser.add_argument('-d', '--keys-data-path', default=None, help='Keys data directory path')
        parser.add_argument('-v', '--model-version-file-path', default=None, help='Model version file path')
        parser.add_argument('-l', '--lookup-package-path', default=None, help='Keys data directory path')
        parser.add_argument('-f', '--keys-format', choices=['oasis', 'json'], help='Keys records / files output format')
        parser.add_argument('-x', '--model-exposures-file-path', default=None, help='Keys records file output format')

    def action(self, args):
        """
        Generate and write Oasis keys (area peril ID, vulnerability ID) for a model.

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)

        lookup_config_fp = as_path(inputs.get('lookup_config_file_path', required=False, is_path=True), 'Lookup config JSON file path',)

        keys_data_path = as_path(inputs.get('keys_data_path', required=False, is_path=True), 'Keys data path', preexists=False)
        model_version_file_path = as_path(inputs.get('model_version_file_path', required=False, is_path=True), 'Model version file path', preexists=False)
        lookup_package_path = as_path(inputs.get('lookup_package_path', required=False, is_path=True), 'Lookup package path', preexists=False)

        if not (lookup_config_fp or (keys_data_path and model_version_file_path and lookup_package_path)):
            raise OasisException('Either the lookup config JSON file path or the keys data path + model version file path + lookup package path must be provided')

        model_exposures_file_path = as_path(inputs.get('model_exposures_file_path', required=True, is_path=True), 'Model exposures')

        keys_format = inputs.get('keys_format', default='oasis')

        self.logger.info('\nGetting model info and lookup')
        model_info, lookup = OasisLookupFactory.create(
            lookup_config_fp=lookup_config_fp,
            model_keys_data_path=keys_data_path,
            model_version_file_path=model_version_file_path,
            lookup_package_path=lookup_package_path
        )
        self.logger.info('\t{}, {}'.format(model_info, lookup))

        utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')

        default_keys_file_name = '{}-{}-{}-keys-{}.{}'.format(model_info['supplier_id'].lower(), model_info['model_id'].lower(), model_info['model_version'], utcnow, 'csv' if keys_format == 'oasis' else 'json')
        default_keys_errors_file_name = '{}-{}-{}-keys-errors-{}.{}'.format(model_info['supplier_id'].lower(), model_info['model_id'].lower(), model_info['model_version'], utcnow, 'csv' if keys_format == 'oasis' else 'json')
           
        keys_file_path = as_path(inputs.get('keys_file_path', default=default_keys_file_name.format(utcnow), required=False, is_path=True), 'Keys file path', preexists=False)
        keys_errors_file_path = as_path(inputs.get('keys_errors_file_path', default=default_keys_errors_file_name.format(utcnow), required=False, is_path=True), 'Keys errors file path', preexists=False)

        self.logger.info('\nSaving keys records to file')
        f1, n1, f2, n2 = OasisLookupFactory.save_results(
            lookup,
            keys_file_path,
            errors_fp=keys_errors_file_path,
            model_exposures_fp=model_exposures_file_path,
            format=keys_format
        )
        self.logger.info('\n{} successful results saved to keys file {}'.format(n1, f1))
        self.logger.info('\n{} unsuccessful results saved to keys errors file {}'.format(n2, f2))


class GenerateOasisFilesCmd(OasisBaseCommand):
    """
    Generate Oasis files (items, coverages, GUL summary) for a model

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    """
    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(self.__class__, self).add_args(parser)

        parser.add_argument('-o', '--oasis-files-path', default=None, help='Path to Oasis files')
        parser.add_argument('-g', '--lookup-config-file-path', default=None, help='Lookup config JSON file path')
        parser.add_argument('-k', '--keys-data-path', default=None, help='Path to Oasis files')
        parser.add_argument('-v', '--model-version-file-path', default=None, help='Model version file path')
        parser.add_argument('-l', '--lookup-package-path', default=None, help='Keys data directory path')
        parser.add_argument(
            '-p', '--canonical-exposures-profile-json-path', default=None,
            help='Path of the supplier canonical exposures profile JSON file'
        )
        parser.add_argument('-e', '--source-exposures-file-path', default=None, help='Source exposures file path')
        parser.add_argument(
            '-a', '--source-exposures-validation-file-path', default=None,
            help='Source exposures validation file (XSD) path'
        )
        parser.add_argument(
            '-b', '--source-to-canonical-exposures-transformation-file-path', default=None,
            help='Source -> canonical exposures transformation file (XSLT) path'
        )
        parser.add_argument(
            '-c', '--canonical-exposures-validation-file-path', default=None,
            help='Canonical exposures validation file (XSD) path'
        )
        parser.add_argument(
            '-d', '--canonical-to-model-exposures-transformation-file-path', default=None,
            help='Canonical exposures validation file (XSD) path'
        )

    def action(self, args):
        """
        Generate Oasis files (items, coverages, GUL summary) for a model

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)

        utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
        default_oasis_files_path = os.path.join(os.getcwd(), 'runs', 'OasisFiles-{}'.format(utcnow))
        oasis_files_path = as_path(inputs.get('oasis_files_path', is_path=True, default=default_oasis_files_path), 'Oasis file', preexists=False)

        lookup_config_fp = as_path(inputs.get('lookup_config_file_path', required=False, is_path=True), 'Lookup config JSON file path', preexists=False)

        keys_data_path = as_path(inputs.get('keys_data_path', required=False, is_path=True), 'Keys data path', preexists=False)
        model_version_file_path = as_path(inputs.get('model_version_file_path', required=False, is_path=True), 'Model version file path', preexists=False)
        lookup_package_path = as_path(inputs.get('lookup_package_path', required=False, is_path=True), 'Lookup package path', preexists=False)

        if not (lookup_config_fp or (keys_data_path and model_version_file_path and lookup_package_path)):
            raise OasisException('Either the lookup config JSON file path or the keys data path + model version file path + lookup package path must be provided')

        canonical_exposures_profile_json_path = as_path(
            inputs.get('canonical_exposures_profile_json_path', required=True, is_path=True),
            'Canonical exposures profile json'
        )
        source_exposures_file_path = as_path(inputs.get('source_exposures_file_path', required=True, is_path=True), 'Source exposures')
        source_exposures_validation_file_path = as_path(
            inputs.get('source_exposures_validation_file_path', required=True, is_path=True),
            'Source exposures validation file'
        )
        source_to_canonical_exposures_transformation_file_path = as_path(
            inputs.get('source_to_canonical_exposures_transformation_file_path', required=True, is_path=True),
            'Source to canonical exposures transformation'
        )
        canonical_exposures_validation_file_path = as_path(
            inputs.get('canonical_exposures_validation_file_path', required=True, is_path=True),
            'Canonical exposures validation file'
        )
        canonical_to_model_exposures_transformation_file_path = as_path(
            inputs.get('canonical_to_model_exposures_transformation_file_path', required=True, is_path=True),
            'Canonical to model exposures transformation file'
        )

        self.logger.info('\nGetting model info and lookup')
        model_info, lookup = OasisLookupFactory.create(
                lookup_config_fp=lookup_config_fp,
                model_keys_data_path=keys_data_path,
                model_version_file_path=model_version_file_path,
                lookup_package_path=lookup_package_path
        )
        self.logger.info('\t{}, {}'.format(model_info, lookup))

        self.logger.info('\nCreating Oasis model object')
        model = OasisExposuresManager().create(
            model_supplier_id=model_info['supplier_id'],
            model_id=model_info['model_id'],
            model_version=model_info['model_version'],
            resources={
                'lookup': lookup,
                'lookup_config_fp': lookup_config_fp or None,
                'oasis_files_path': oasis_files_path,
                'source_exposures_file_path': source_exposures_file_path,
                'source_exposures_validation_file_path': source_exposures_validation_file_path,
                'source_to_canonical_exposures_transformation_file_path': source_to_canonical_exposures_transformation_file_path,
                'canonical_exposures_profile_json_path': canonical_exposures_profile_json_path,
                'canonical_exposures_validation_file_path': canonical_exposures_validation_file_path,
                'canonical_to_model_exposures_transformation_file_path': canonical_to_model_exposures_transformation_file_path
            }
        )
        self.logger.info('\t{}'.format(model))

        self.logger.info('\nSetting up Oasis files directory for model {}'.format(model.key))
        Path(oasis_files_path).mkdir(parents=True, exist_ok=True)

        self.logger.info('\nGenerating Oasis files for model')
        oasis_files = OasisExposuresManager().start_files_pipeline(
            oasis_model=model,
            logger=self.logger,
        )

        self.logger.info('\nGenerated Oasis files for model: {}'.format(oasis_files))


class GenerateLossesCmd(OasisBaseCommand):
    """
    Generate losses using the installed ktools framework.

    Given Oasis files, model analysis settings JSON file, model data, and
    some other parameters. can generate losses using the installed ktools framework.

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    Run ``oasislmf config --help`` for more information.

    The script creates a time-stamped folder in the model run directory and
    sets that as the new model run directory, copies the analysis settings
    JSON file into the run directory and creates the following folder
    structure
    ::

        ├── analysis_settings.json
        ├── fifo/
        ├── input/
        ├── output/
        ├── static/
        └── work/

    Depending on the OS type the model data is symlinked (Linux, Darwin) or
    copied (Cygwin, Windows) into the ``static`` subfolder. The input files
    are kept in the ``input`` subfolder and the losses are generated as CSV
    files in the ``output`` subfolder.

    By default executing the generated ktools losses script will automatically
    execute, this can be overridden by providing the ``--no-execute`` flag.
    """
    formatter_class = RawDescriptionHelpFormatter

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(self.__class__, self).add_args(parser)

        parser.add_argument('-o', '--oasis-files-path', default=None, help='Path to Oasis files')
        parser.add_argument('-j', '--analysis-settings-json-file-path', default=None, help='Relative or absolute path of the model analysis settings JSON file')
        parser.add_argument('-m', '--model-data-path', default=None, help='Model data source path')
        parser.add_argument('-r', '--model-run-dir-path', default=None, help='Model run directory path')
        parser.add_argument('-s', '--ktools-script-name', default=None, help='Relative or absolute path of the output file')
        parser.add_argument('-n', '--ktools-num-processes', default=-1, help='Number of ktools calculation processes to use')
        parser.add_argument('-x', '--no-execute', action='store_true', help='Whether to execute generated ktools script')

    def action(self, args):
        """
        Generate losses using the installed ktools framework.

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)

        oasis_files_path = as_path(inputs.get('oasis_files_path', required=True, is_path=True), 'Oasis files', preexists=True)

        model_run_dir_path = as_path(inputs.get('model_run_dir_path', required=False, is_path=True), 'Model run directory', preexists=False)

        analysis_settings_json_file_path = as_path(
            inputs.get('analysis_settings_json_file_path', required=True, is_path=True),
            'Analysis settings file'
        )
        model_data_path = as_path(inputs.get('model_data_path', required=True, is_path=True), 'Model data')

        ktools_script_name = inputs.get('ktools_script_name', default='run_ktools')
        no_execute = inputs.get('no_execute', default=False)

        if not model_run_dir_path:
            utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
            model_run_dir_path = os.path.join(os.getcwd(), 'runs', 'ProgOasis-{}'.format(utcnow))
            self.logger.info('\nNo model run dir. provided - creating a timestamped run dir. in working directory as {}'.format(model_run_dir_path))
            Path(model_run_dir_path).mkdir(parents=True, exist_ok=True)
        else:
            if not os.path.exists(model_run_dir_path):
                Path(model_run_dir_path).mkdir(parents=True, exist_ok=True)

        self.logger.info(
            '\nPreparing model run directory {} - copying Oasis files, analysis settings JSON file and linking model data'.format(model_run_dir_path)
        )
        prepare_model_run_directory(
            model_run_dir_path,
            oasis_files_path,
            analysis_settings_json_file_path,
            model_data_path
        )

        self.logger.info('\nConverting Oasis files to ktools binary files')
        oasis_files_path = os.path.join(model_run_dir_path, 'input', 'csv')
        binary_files_path = os.path.join(model_run_dir_path, 'input')
        create_binary_files(oasis_files_path, binary_files_path)

        analysis_settings_json_file_path = os.path.join(model_run_dir_path, 'analysis_settings.json')
        try:
            self.logger.info('\nReading analysis settings JSON file')
            with io.open(analysis_settings_json_file_path, 'r', encoding='utf-8') as f:
                analysis_settings = json.load(f)
                if 'analysis_settings' in analysis_settings:
                    analysis_settings = analysis_settings['analysis_settings']
        except (IOError, TypeError, ValueError):
            raise OasisException('Invalid analysis settings JSON file or file path: {}.'.format(analysis_settings_json_file_path))

        self.logger.info('\nLoaded analysis settings JSON: {}'.format(analysis_settings))

        self.logger.info('\nPreparing model run inputs')
        prepare_model_run_inputs(analysis_settings, model_run_dir_path)

        script_path = os.path.join(model_run_dir_path, '{}.sh'.format(ktools_script_name))
        if no_execute:
            self.logger.info('\nGenerating ktools losses script')
            genbash(
                args.ktools_num_processes,
                analysis_settings,
                filename=script_path,
            )
            self.logger.info('\nMaking ktools losses script executable')
            subprocess.check_call("chmod +x {}".format(script_path), stderr=subprocess.STDOUT, shell=True)
        else:
            os.chdir(model_run_dir_path)
            run(analysis_settings, args.ktools_num_processes, filename=script_path)

        self.logger.info('\nLoss outputs generated in {}'.format(os.path.join(model_run_dir_path, 'output')))


class RunCmd(OasisBaseCommand):
    """
    Run models end to end.

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    """
    def add_args(self, parser):
        """
        Run models end to end.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(self.__class__, self).add_args(parser)

        parser.add_argument('-k', '--keys-data-path', default=None, help='Path to Oasis files')
        parser.add_argument('-v', '--model-version-file-path', default=None, help='Model version file path')
        parser.add_argument('-l', '--lookup-package-file-path', default=None, help='Keys data directory path')
        parser.add_argument('-g', '--lookup-config-file-path', default=None, help='Lookup config JSON file path')
        parser.add_argument(
            '-p', '--canonical-exposures-profile-json-path', default=None,
            help='Path of the supplier canonical exposures profile JSON file'
        )
        parser.add_argument('-e', '--source-exposures-file-path', default=None, help='Source exposures file path')
        parser.add_argument(
            '-a', '--source-exposures-validation-file-path', default=None,
            help='Source exposures validation file (XSD) path'
        )
        parser.add_argument(
            '-b', '--source-to-canonical-exposures-transformation-file-path', default=None,
            help='Source -> canonical exposures transformation file (XSLT) path'
        )
        parser.add_argument(
            '-c', '--canonical-exposures-validation-file-path', default=None,
            help='Canonical exposures validation file (XSD) path'
        )
        parser.add_argument(
            '-d', '--canonical-to-model-exposures-transformation-file-path', default=None,
            help='Canonical exposures validation file (XSD) path'
        )
        parser.add_argument(
            '-j', '--analysis-settings-json-file-path', default=None,
            help='Model analysis settings JSON file path'
        )
        parser.add_argument('-m', '--model-data-path', default=None, help='Model data source path')
        parser.add_argument('-r', '--model-run-dir-path', default=None, help='Model run directory path')
        parser.add_argument(
            '-s', '--ktools-script-name', default=None,
            help='Name of the ktools output script (should not contain any filetype extension)'
        )
        parser.add_argument('-n', '--ktools-num-processes', default=2, help='Number of ktools calculation processes to use')

    def action(self, args):
        """
        Generate Oasis files (items, coverages, GUL summary) for a model

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)
        model_run_dir_path = as_path(inputs.get('model_run_dir_path', required=False), 'Model run path', preexists=False)

        if not model_run_dir_path:
            utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
            model_run_dir_path = os.path.join(os.getcwd(), 'runs', 'ProgOasis-{}'.format(utcnow))
            self.logger.info('\nNo model run dir. provided - creating a timestamped run dir. in working directory as {}'.format(model_run_dir_path))
            Path(model_run_dir_path).mkdir(parents=True, exist_ok=True)
        else:
            if not os.path.exists(model_run_dir_path):
                Path(model_run_dir_path).mkdir(parents=True, exist_ok=True)

        args.model_run_dir_path = model_run_dir_path

        args.oasis_files_path = os.path.join(model_run_dir_path, 'tmp')
        self.logger.info('\nCreating temporary folder {} for Oasis files'.format(args.oasis_files_path))
        Path(args.oasis_files_path).mkdir(parents=True, exist_ok=True)

        gen_oasis_files_cmd = GenerateOasisFilesCmd()
        gen_oasis_files_cmd._logger = self.logger
        gen_oasis_files_cmd.action(args)

        gen_losses_cmd = GenerateLossesCmd()
        gen_losses_cmd._logger = self.logger
        gen_losses_cmd.action(args)


class ModelsCmd(OasisBaseCommand):
    sub_commands = {
        'generate-peril-areas-rtree-file-index': GeneratePerilAreasRtreeFileIndexCmd,
        'transform-source-to-canonical': TransformSourceToCanonicalFileCmd,
        'transform-canonical-to-model': TransformCanonicalToModelFileCmd,
        'generate-keys': GenerateKeysCmd,
        'generate-oasis-files': GenerateOasisFilesCmd,
        'generate-losses': GenerateLossesCmd,
        'run': RunCmd,
    }



# -*- coding: utf-8 -*-

__all__ = [
    'DEFAULT_RTREE_INDEX_PROPS',
    'generate_index_entries',
    'get_peril_areas',
    'get_peril_areas_index',
    'get_rtree_index',
    'PerilArea',
    'PerilAreasIndex',
    'PERIL_ID_FLOOD',
    'PERIL_ID_QUAKE',
    'PERIL_ID_SURGE',
    'PERIL_ID_WIND'
]

import builtins
import copy
import io
import json
import os
import re
import types
import uuid

from collections import OrderedDict

import rtree

from rtree.index import (
    Index as RTreeIndex,
    Property as RTreeIndexProperty,
)

from rtree.core import RTreeError

from shapely import speedups as shapely_speedups

if shapely_speedups.available:
    shapely_speedups.enable()

from shapely.geometry import (
    box,
    Point,
    MultiPoint,
    Polygon,
)

import six

from six.moves import cPickle as cpickle

from .exceptions import OasisException
from .data import get_dataframe


PERIL_ID_WIND = 1
PERIL_ID_SURGE = 2
PERIL_ID_QUAKE = 3
PERIL_ID_FLOOD = 4

DEFAULT_RTREE_INDEX_PROPS = {
    'buffering_capacity': 10,
    'custom_storage_callbacks': None,
    'custom_storage_callbacks_size': 0,
    'dat_extension': 'dat',
    'dimension': 2,
    'filename': '',
    'fill_factor': 0.7,
    'idx_extension': 'idx',
    'index_capacity': 100,
    'index_id': None,
    'leaf_capacity': 100,
    'near_minimum_overlap_factor': 32,
    'overwrite': True,
    'pagesize': 4096,
    'point_pool_capacity': 500,
    'region_pool_capacity': 1000,
    'reinsert_factor': 0.3,
    'split_distribution_factor': 0.4,
    'storage': 0,
    'tight_mbr': True,
    'tpr_horizon': 20.0,
    'type': 0,
    'variant': 2,
    'writethrough': False
}


def generate_index_entries(items, objects=None):
    if objects:
        for (key, poly_bounds), obj in zip(items, objects):
            yield key, poly_bounds, obj
    else:
        for key, poly_bounds in items:
            yield key, poly_bounds, None


def get_peril_areas(areas):
    for peril_id, coverage_type, peril_area_id, coordinates, other_props in areas:
        yield PerilArea(coordinates, peril_id=peril_id, coverage_type=coverage_type, peril_area_id=peril_area_id, **other_props)


def get_peril_areas_index(
    areas=None,
    peril_areas=None,
    properties=None
):
    if not (areas or peril_areas):
        raise OasisException('Either areas or peril areas must be provided')

    return PerilAreasIndex(areas=areas, peril_areas=peril_areas, properties=properties)


def get_rtree_index(
    items,
    objects=None,
    properties=None
):
    return (
        RTreeIndex(generate_index_entries(items, objects=objects), properties=RTreeIndexProperty(**properties)) if properties
        else RTreeIndex(generate_index_entries(items, objects=objects))
    )


class PerilArea(Polygon):

    def __init__(self, coords, **kwargs):

        _coords = tuple(c for c in coords)

        if not _coords:
            raise OasisException('No peril area coordinates')

        if len(_coords) > 2:
            self._multipoint = MultiPoint(_coords)
        elif len(_coords) == 2:
            minx, miny, maxx, maxy = tuple(_c for c in _coords for _c in c)
            self._multipoint = MultiPoint(box(minx, miny, maxx, maxy).exterior.coords)
        elif len(_coords) == 1:
            x, y = _coords[0][0], _coords[0][1]
            r = kwargs.get('area_reg_poly_radius') or 0.0016
            self._multipoint = MultiPoint(
                tuple((x + r*(-1)**i, y + r*(-1)**j) for i in range(2) for j in range(2))
            )

        super(self.__class__, self).__init__(shell=self._multipoint.convex_hull.exterior.coords)
        
        self._coordinates = tuple(self.exterior.coords)

        self._centre = self.centroid.x, self.centroid.y

        self._coverage_type = kwargs.get('coverage_type')

        self._peril_id = kwargs.get('peril_id')

        self._id = kwargs.get('area_peril_id') or kwargs.get('peril_area_id') or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

    @property
    def multipoint(self):
        return self._multipoint

    @property
    def coordinates(self):
        return self._coordinates

    @property
    def centre(self):
        return self._centre

    @property
    def coverage_type(self):
        return self._coverage_type

    @property
    def peril_id(self):
        return self._peril_id

    @property
    def id(self):
        return self._id


class PerilAreasIndex(RTreeIndex):

    def __init__(self, *args, **kwargs):

            self._protocol = (2 if six.sys.version_info[0] < 3 else cpickle.HIGHEST_PROTOCOL)

            idx_fp = kwargs.get('fp')

            areas = kwargs.get('areas')
            peril_areas = kwargs.get('peril_areas')

            props = kwargs.get('properties') or copy.deepcopy(DEFAULT_RTREE_INDEX_PROPS)

            if not (idx_fp or areas or peril_areas):
                self._peril_areas = self._stream = None
                kwargs['properties'] = RTreeIndexProperty(**props)
                super(self.__class__, self).__init__(*args, **kwargs)
            elif idx_fp:
                self._peril_areas = self._stream = None
                _idx_fp = idx_fp
                if not os.path.isabs(_idx_fp):
                    _idx_fp = os.path.abspath(_idx_fp)
                
                idx_ext = props.get('idx_extension') or 'idx'
                dat_ext = props.get('dat_extension') or 'dat'

                if not (os.path.exists('{}.{}'.format(_idx_fp, idx_ext)) or os.path.exists('{}.{}'.format(_idx_fp, dat_ext))):
                    kwargs['properties'] = RTreeIndexProperty(**props)

                super(self.__class__, self).__init__(_idx_fp, *args, **kwargs)
            else:
                self._peril_areas = OrderedDict({
                    pa.id:pa for pa in (peril_areas if peril_areas else self._get_peril_areas(areas))
                })
                self._stream = self._generate_index_entries(
                    ((paid, pa.bounds) for paid, pa in six.iteritems(self._peril_areas)),
                    objects=((paid, pa.bounds, pa.coordinates) for paid, pa in six.iteritems(self._peril_areas))
                )
                kwargs['properties'] = RTreeIndexProperty(**index_props)
                super(self.__class__, self).__init__(self._stream, *args, **kwargs)

    def dumps(self, obj):
        return cpickle.dumps(obj, protocol=self.protocol)

    def loads(self, data):
        return cpickle.loads(data)

    def _get_peril_areas(self, areas):
        for peril_id, coverage_type, peril_area_id, coordinates, other_props in areas:
            yield PerilArea(coordinates, peril_id=peril_id, coverage_type=coverage_type, peril_area_id=peril_area_id, **other_props)

    def _generate_index_entries(self, items, objects=None):
        if objects:
            for (key, poly_bounds), obj in zip(items, objects):
                yield key, poly_bounds, obj
        else:
            for key, poly_bounds in items:
                yield key, poly_bounds, None

    @property
    def protocol(self):
        return self._protocol
    

    @property
    def peril_areas(self):
        return self._peril_areas

    @property
    def stream(self):
        if self._peril_areas:
            self._stream = self._generate_index_entries(self._peril_areas)
            return self._stream
        return None

    @classmethod
    def create_from_peril_areas_file(
        cls,
        src_fp=None,
        src_type='csv',
        peril_id_col='peril_id',
        coverage_type_col='coverage_type',
        peril_area_id_col='area_peril_id',
        non_na_cols=('peril_id', 'coverage_type', 'area_peril_id',),
        col_dtypes={'peril_id': int, 'coverage_type': int, 'area_peril_id': int},
        sort_col='area_peril_id',
        area_poly_coords_cols={},
        area_poly_coords_seq_start_idx=1,
        area_reg_poly_radius=0.00166,
        static_props={},
        index_fp=None,
        index_props=copy.deepcopy(DEFAULT_RTREE_INDEX_PROPS)
    ):
        if not src_fp:
            raise OasisException(
                'An areas source CSV or JSON file path must be provided'
            )

        _src_fp = src_fp
        if not os.path.isabs(_src_fp):
            _src_fp = os.path.abspath(_src_fp)

        _non_na_cols = set(non_na_cols)

        _peril_id_col = peril_id_col.lower()
        _coverage_type_col = coverage_type_col.lower()
        _peril_area_id_col = peril_area_id_col.lower()

        if not set(_non_na_cols).intersection([_peril_id_col, _coverage_type_col, _peril_area_id_col]):
            _non_na_cols = _non_na_cols.union({_peril_id_col, _coverage_type_col, _peril_area_id_col})

        for col in six.itervalues(area_poly_coords_cols):
            if not col in _non_na_cols:
                _non_na_cols = _non_na_cols.union({col.lower()})

        _non_na_cols = tuple(_non_na_cols)

        _sort_col = sort_col.lower()

        areas_df = get_dataframe(
            src_fp=_src_fp,
            src_type=src_type,
            non_na_cols=_non_na_cols,
            col_dtypes=col_dtypes,
            sort_col=(_peril_area_id_col or _sort_col)
        )

        coords_cols = area_poly_coords_cols

        seq_start = area_poly_coords_seq_start_idx

        len_seq = sum(1 if re.match(r'x(\d+)?', k) else 0 for k in six.iterkeys(coords_cols))

        peril_areas = cls()._get_peril_areas(
            (
                ar[_peril_id_col],
                ar[_coverage_type_col],
                ar[_peril_area_id_col],
                tuple(
                    (ar.get(coords_cols['x{}'.format(i)].lower()) or 0, ar.get(coords_cols['y{}'.format(i)].lower()) or 0)
                    for i in range(seq_start, len_seq + 1)
                ),
                static_props
            ) for _, ar in areas_df.iterrows()
        )

        _index_fp = index_fp
        if not _index_fp:
            raise OasisException('No output file index path provided')

        if not os.path.isabs(_index_fp):
            _index_fp = os.path.abspath(_index_fp)

        try:
            return cls().save(
                _index_fp,
                peril_areas=peril_areas,
                index_props=index_props
            )
        except OasisException as e:
            raise

    def save(
        self,
        index_fp,
        peril_areas=None,
        index_props=DEFAULT_RTREE_INDEX_PROPS
    ):
        _index_fp = index_fp

        if not os.path.isabs(_index_fp):
            _index_fp = os.path.abspath(_index_fp)

        if os.path.exists(_index_fp):
            os.remove(_index_fp)

        class myindex(RTreeIndex):
            def __init__(self, *args, **kwargs):
                self.protocol = (2 if six.sys.version_info[0] < 3 else cpickle.HIGHEST_PROTOCOL)
                super(self.__class__, self).__init__(*args, **kwargs)

            def dumps(self, obj):
                return cpickle.dumps(obj, protocol=self.protocol)

            def loads(self, obj):
                return cpickle.loads(obj)

        try:
            index = myindex(_index_fp, properties=RTreeIndexProperty(**index_props))

            _peril_areas = self._peril_areas or peril_areas

            if not _peril_areas:
                raise OasisException(
                    'No peril areas found in instance or in arguments - '
                    'this is required to write the index to file'
                )

            peril_areas_seq = None

            if (isinstance(peril_areas, list) or isinstance(peril_areas, tuple)):
                peril_areas_seq = (pa for pa in peril_areas)
            elif isinstance(peril_areas, types.GeneratorType):
                peril_areas_seq = peril_areas
            elif (isinstance(peril_areas, dict)):
                peril_areas_seq = six.itervalues(peril_areas)

            for pa in peril_areas_seq:
                index.insert(pa.id, pa.bounds, obj=(pa.peril_id, pa.coverage_type, pa.id, pa.bounds, pa.coordinates))

            index.close()
        except (IOError, OSError, RTreeError) as e:
            raise

        return _index_fp



from argparse import RawDescriptionHelpFormatter

from .base import OasisBaseCommand


class ConfigCmd(OasisBaseCommand):
    """
    The ``oasislmf`` tool can be configured using using a JSON configuration file. 
    By default this is stored in ``oasislmf.json`` but the path can be overridden using the ``--config`` flag.

    This configuration file should contain any or all of the following properties.
    The path-related keys should be string paths, given relative to the location of JSON file.
    Arguments can be overridden for a specific command using appropriate run time parameter.

    :source_exposures_file_path: File path to source exposure data, in CSV format.
    :source_exposures_validation_file_path: File path to schema for validating source exposure, in XSD format.
    :source_to_canonical_exposures_transformation_file_path: File path to transform to select and store a subset of the source exposure, in XSLT format.
    :canonical_exposures_profile_json_path: File path to profile that describes the structure of the stored data.
    :canonical_exposures_validation_file_path: File path to schema for validating stored exposure, in XSD format.
    :canonical_to_model_exposures_transformation_file_path: File path to transform to select a subset of the stored exposure for invoking the keys lookup logic, in XSD format.
    :lookup_package_path: Path to the Python package that contains the model keys lookup class.
    :keys_data_path: Path to directory containing data files required by the model keys lookup class.
    :model_version_file_path: Path to model version specification file (`ModelVersion.csv`), required to correctly instantiate the model keys lookup class.
    :model_exposures_file_path: File path to the inputs to the model keys lookup class in CSV format.
    :keys_file_path: Path for the file containing keys for locations with successful lookups.
    :keys_error_file_path: Path for the file containing keys for locations with failing or non-matching looups.
    :keys_format: Specifies the format of the keys files generated by the model keys lookup class. The default is `oasis` (Oasis CSV format), with `json` as alternative choice.
    :model_data_path: Path to directory containing the model data.
    :model_run_dir_path: Path to directory in which the canonical files, keys files, Oasis files and output files are generated, including the source files and analysis settings file, which are copied from their original locations.
    :analysis_settings_json_file_path: File path to the settings file that parameterizes the model run.
    :ktools_script_name: Specifies the ktools shell script that performs the model run, which is created and ran from the model_run_dir_path. Default is "run_ktools.sh".
    :ktools_num_processes: Specifies the number of concurrent processes used by the model run. Default is 2. 
    :no_execute: Specifies whether to perform the model run. Default is "False", which means the ktools script is executed.


    As an example, this is the master script configuration file for PiWind

    ::

        {
            "source_exposures_file_path": "tests/data/SourceLocPiWind1K.csv",
            "source_exposures_validation_file_path": "flamingo/PiWind/Files/ValidationFiles/Generic_Windstorm_SourceLoc.xsd",
            "source_to_canonical_exposures_transformation_file_path": "flamingo/PiWind/Files/TransformationFiles/MappingMapToGeneric_Windstorm_CanLoc_A.xslt",
            "canonical_exposures_profile_json_path": "oasislmf-piwind-canonical-loc-profile.json",
            "canonical_exposures_validation_file_path": "flamingo/PiWind/Files/ValidationFiles/Generic_Windstorm_CanLoc_B.xsd",
            "canonical_to_model_exposures_transformation_file_path": "flamingo/PiWind/Files/TransformationFiles/MappingMapTopiwind_modelloc.xslt",
            "lookup_package_path": "src/keys_server",
            "keys_data_path": "keys_data/PiWind",
            "model_version_file_path": "keys_data/PiWind/ModelVersion.csv",
            "model_exposures_file_path": "/tmp/model_exposure_PiWind.csv",
            "keys_file_path": "/tmp/keys_PiWind.csv",
            "keys_error_file_path": "/tmp/keys_erros_PiWind.json",
            "model_data_path": "model_data/PiWind",
            "model_run_dir_path": "/tmp/run_PiWind",
            "analysis_settings_json_file_path": "analysis_settings.json"
        }

    It can also be obtained `here <https://github.com/OasisLMF/OasisPiWind/blob/master/oasislmf.json>`_.

   """
    formatter_class = RawDescriptionHelpFormatter



import filecmp
import os
import unittest


# find the root of git repo & import class under test
# Set Dir Vars
from backports.tempfile import TemporaryDirectory
from hypothesis import given
from hypothesis.strategies import integers
from pathlib2 import Path

from oasislmf.exposures.csv_trans import Translator
from oasislmf.utils.diff import unified_diff

data_dir = str(Path(__file__).parent.joinpath('csv_trans_data'))
input_data_dir = str(Path(data_dir, 'input'))
expected_data_dir = str(Path(data_dir, 'expected'))


class CsvTrans(unittest.TestCase):
    @given(integers(min_value=1, max_value=10))
    def test_source_to_canonical(self, chunk_size):
        with TemporaryDirectory() as d:
            output_file = os.path.join(d, 'canonical.csv')

            translator = Translator(
                os.path.join(input_data_dir, 'source.csv'),
                output_file,
                os.path.join(input_data_dir, 'source_to_canonical.xsd'),
                os.path.join(input_data_dir, 'source_to_canonical.xslt'),
                chunk_size=chunk_size,
                append_row_nums=True
            )
            translator()

            diff = unified_diff(output_file, os.path.join(expected_data_dir, 'canonical.csv'), as_string=True)
            self.assertEqual(0, len(diff), diff)

    @given(integers(min_value=1, max_value=10))
    def test_canonical_to_model(self, chunk_size):
        with TemporaryDirectory() as d:
            output_file = os.path.join(d, 'model.csv')

            translator = Translator(
                os.path.join(input_data_dir, 'canonical.csv'),
                output_file,
                os.path.join(input_data_dir, 'canonical_to_model.xsd'),
                os.path.join(input_data_dir, 'canonical_to_model.xslt'),
                chunk_size=chunk_size
            )
            translator()
            self.assertTrue(filecmp.cmp(output_file, os.path.join(expected_data_dir, 'model.csv')))



import logging
from logging.handlers import RotatingFileHandler
from random import random
from unittest import TestCase

from hypothesis import given
from hypothesis.strategies import sampled_from
from mock import Mock, patch

from oasislmf.utils.log import (
    oasis_log,
    read_log_config,
    set_rotating_logger,
)


class MockLogger(object):
    def __init__(self):
        self.debug = Mock()
        self.info = Mock()


def create_callable(result=None):
    def _callable(*args, **kwargs):
        if not hasattr(_callable, 'calls'):
            _callable.calls = []

        _callable.calls.append({'args': args, 'kwargs': kwargs})
        return result

    return _callable


class OasisLog(TestCase):
    def test_function_is_wrapped_without_arg_list___return_value_is_correct(self):
        expected = random()
        callable = create_callable(result=expected)

        wrapped = oasis_log(callable)
        res = wrapped()

        self.assertEqual(expected, res)

    def test_function_is_wrapped_with_arg_list___return_value_is_correct(self):
        expected = random()
        callable = create_callable(result=expected)

        wrapped = oasis_log()(callable)
        res = wrapped()

        self.assertEqual(expected, res)

    def test_function_is_wrapped_without_arg_list___callable_is_called_once_with_the_correct_args(self):
        callable = create_callable()

        wrapped = oasis_log(callable)
        wrapped('first', second='second')

        self.assertEqual(1, len(callable.calls))
        self.assertEqual(('first', ), callable.calls[0]['args'])
        self.assertEqual({'second': 'second'}, callable.calls[0]['kwargs'])

    def test_function_is_wrapped_with_arg_list___callable_is_called_once_with_the_correct_args(self):
        callable = create_callable()

        wrapped = oasis_log()(callable)
        wrapped('first', second='second')

        self.assertEqual(1, len(callable.calls))
        self.assertEqual(('first', ), callable.calls[0]['args'])
        self.assertEqual({'second': 'second'}, callable.calls[0]['kwargs'])

    def test_wrapped_function_is_called___info_logging_is_called(self):
        logger_mock = MockLogger()

        with patch('oasislmf.utils.log.logging.getLogger', Mock(return_value=logger_mock)):
            callable = create_callable()

            wrapped = oasis_log(callable)
            wrapped('first', second='second')

            self.assertGreater(logger_mock.info.call_count, 0)

    def test_wrapped_function_is_called___args_and_kwargs_are_logged_to_debug_excluding_self(self):
        logger_mock = MockLogger()

        with patch('oasislmf.utils.log.logging.getLogger', Mock(return_value=logger_mock)):
            class FakeObj(object):
                @oasis_log
                def method(self, first, second=None):
                    pass

            FakeObj().method('a', second='b')

            logger_mock.debug.assert_any_call("    first == a")
            logger_mock.debug.assert_any_call("    second == b")

            for args in logger_mock.debug.call_args_list:
                if 'self' in args[0][0]:
                    self.fail('"self" was logged')


class ReadLogConfig(TestCase):
    @given(sampled_from([logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]))
    def test_log_config_is_loaded___logger_is_updated(self, level):
        with patch('logging.root', logging.RootLogger(logging.NOTSET)):
            read_log_config({
                'LOG_FILE': '/tmp/log_file.txt',
                'LOG_LEVEL': level,
                'LOG_MAX_SIZE_IN_BYTES': 100,
                'LOG_BACKUP_COUNT': 10,
            })

            logger = logging.getLogger()

            self.assertEqual(level, logger.level)

            self.assertEqual(1, len(logger.handlers))
            handler = logger.handlers[0]
            self.assertIsInstance(handler, RotatingFileHandler)
            self.assertEqual('/tmp/log_file.txt', handler.baseFilename)
            self.assertEqual(100, handler.maxBytes)
            self.assertEqual(10, handler.backupCount)


class SetRotatingLogger(TestCase):
    @given(
        level=sampled_from([logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG])
    )
    def test_rotating_log_config_is_loaded___logger_is_set(self, level):
        with patch('logging.root', logging.RootLogger(logging.NOTSET)):
            set_rotating_logger(log_file_path='/tmp/log_file.txt', log_level=level, max_file_size=100, max_backups=10)

            logger = logging.getLogger()

            self.assertEqual(level, logger.level)

            self.assertEqual(1, len(logger.handlers))
            handler = logger.handlers[0]
            self.assertIsInstance(handler, RotatingFileHandler)
            self.assertEqual('/tmp/log_file.txt', handler.baseFilename)
            self.assertEqual(100, handler.maxBytes)
            self.assertEqual(10, handler.backupCount)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

__all__ = [
    'OasisBaseLookup',
    'OasisBaseKeysLookup',
    'OasisLookup',
    'OasisPerilLookup',
    'OasisVulnerabilityLookup',
    'OasisLookupFactory'
]

import builtins
import csv
import imp
import importlib
import io
import itertools
import json
import os
import re
import sys
import types
import uuid

from collections import OrderedDict

import pandas as pd
import six

from shapely.geometry import (
    box,
    Point,
)

from shapely import speedups as shapely_speedups

if shapely_speedups.available:
    shapely_speedups.enable()

from rtree.core import RTreeError

import six

from ..utils.data import get_dataframe
from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from ..utils.peril import (
    DEFAULT_RTREE_INDEX_PROPS,
    PerilAreasIndex,
)
from ..utils.status import (
    KEYS_STATUS_FAIL,
    KEYS_STATUS_NOMATCH,
    KEYS_STATUS_SUCCESS,
)
from ..utils.values import is_string


UNKNOWN_ID = -1

def as_path(value, name, preexists=True):
    """
    Processes the path and returns the absolute path.

    If the path does not exist and ``preexists`` is true
    an ``OasisException`` is raised.

    :param value: The path to process
    :type value: str

    :param name: The name of the path (used for error reporting)
    :type name: str

    :param preexists: Flag whether to raise an error if the path
        does not exist.
    :type preexists: bool

    :return: The absolute path of the input path
    """
    if value is not None:
        value = os.path.abspath(value) if not os.path.isabs(value) else value

    if preexists and not (value is not None and os.path.exists(value)):
        raise OasisException('{} does not exist: {}'.format(name, value))

    return value


class OasisBaseLookup(object):

    @oasis_log()
    def __init__(self, config=None, config_json=None, config_fp=None):
        if config:
            self._config = config
        elif config_json:
            self._config = json.loads(config_json)
        elif config_fp:
            _config_fp = as_path(config_fp, 'config_fp')
            with io.open(_config_fp, 'r', encoding='utf-8') as f:
                self._config = json.load(f)

        keys_data_path = self._config.get('keys_data_path') or ''

        self._config['keys_data_path'] = as_path(keys_data_path, 'keys_data_path', preexists=(True if keys_data_path else False))

        peril_config = self._config.get('peril') or {}

        self._peril_ids = tuple(peril_config.get('peril_ids') or ())

        self._peril_id_col = peril_config.get('peril_id_col') or 'peril_id'

        coverage_config = self._config.get('coverage') or {}

        self._coverage_types = tuple(coverage_config.get('coverage_types') or ())

        self._coverage_type_col = peril_config.get('coverage_type_col') or 'coverage_type'

        self.__tweak_config_data__()

    def __tweak_config_data__(self):
        for section in ('locations', 'peril', 'vulnerability',):
            section_config = self._config.get(section) or {}
            for k, v in six.iteritems(section_config):
                if is_string(v) and '%%KEYS_DATA_PATH%%' in v:
                    self._config[section][k] = v.replace('%%KEYS_DATA_PATH%%', self._config['keys_data_path'])
                elif type(v) == list:
                    self._config[section][k] = tuple(v)
                elif isinstance(v, dict):
                    for _k, _v in six.iteritems(v):
                        if is_string(_v) and '%%KEYS_DATA_PATH%%' in _v:
                            self._config[section][k][_k] = _v.replace('%%KEYS_DATA_PATH%%', self._config['keys_data_path'])

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, c):
        self._config = c
        self.__tweak_config_data__()

    @property
    def peril_ids(self):
        return self._peril_ids

    @property
    def peril_id_col(self):
        return self._peril_id_col

    @property
    def coverage_types(self):
        return self._coverage_types

    @property
    def coverage_type_col(self):
        return self._coverage_type_col

    def lookup(self, loc, peril_id, coverage_type, **kwargs):
        """
        Lookup for an individual location item, which could be a dict or a
        Pandas series object.
        """
        pass

    @oasis_log()
    def bulk_lookup(self, locs, **kwargs):
        """
        Bulk vulnerability lookup for a list, tuple, generator, pandas data
        frame or dict of location items, which can be dicts or Pandas series
        objects or any object which has as a dict-like interface.

        Generates results using ``yield``.
        """
        locs_seq = None

        if (isinstance(locs, list) or isinstance(locs, tuple)):
            locs_seq = (loc for loc in locs)
        elif isinstance(locs, types.GeneratorType):
            locs_seq = locs
        elif (isinstance(locs, dict)):
            locs_seq = six.itervalues(locs)
        elif isinstance(locs, pd.DataFrame):
            locs_seq = (loc for _, loc in locs.iterrows())

        for peril_id, coverage_type, loc in itertools.product(self.peril_ids, self.coverage_types, locs_seq):
            yield self.lookup(loc, peril_id, coverage_type)


class OasisBaseKeysLookup(object):  # pragma: no cover
    """
    A base class / interface that serves a template for model-specific keys
    lookup classes.
    """
    @oasis_log()
    def __init__(
        self,
        keys_data_directory=None,
        supplier=None,
        model_name=None,
        model_version=None
    ):
        """
        Class constructor
        """
        if keys_data_directory is not None:
            self.keys_data_directory = keys_data_directory
        else:
            self.keys_data_directory = os.path.join(os.sep, 'var', 'oasis', 'keys_data')

        self.supplier = supplier
        self.model_name = model_name
        self.model_version = model_version

    @oasis_log()
    def process_locations(self, loc_df):
        """
        Process location rows - passed in as a pandas dataframe.
        """
        pass

    def _get_location_record(self, raw_loc_item):
        """
        Returns a dict of standard location keys and values based on
        a raw location item, which is a row in a Pandas dataframe.
        """
        pass

    def _get_area_peril_id(self, record):
        """
        Get the area peril ID for a particular location record.
        """
        return UNKNOWN_ID, "Not implemented"

    def _get_vulnerability_id(self, record):
        """
        Get the vulnerability ID for a particular location record.
        """
        return UNKNOWN_ID, "Not implemented"

    @oasis_log()
    def _get_area_peril_ids(self, loc_data, include_context=True):
        """
        Generates area peril IDs in two modes - if include_context is
        True (default) it will generate location records/rows including
        the area peril IDs, otherwise it will generate pairs of location
        IDs and the corresponding area peril IDs.
        """
        pass

    @oasis_log()
    def _get_vulnerability_ids(self, loc_data, include_context=True):
        """
        Generates vulnerability IDs in two modes - if include_context is
        True (default) it will generate location records/rows including
        the area peril IDs, otherwise it will generate pairs of location
        IDs and the corresponding vulnerability IDs.
        """
        pass

    def _get_lookup_success(self, ap_id, vul_id):
        """
        Determine the status of the keys lookup.
        """
        if ap_id == UNKNOWN_ID or vul_id == UNKNOWN_ID:
            return KEYS_STATUS_NOMATCH
        return KEYS_STATUS_SUCCESS


class OasisLookupFactory(object):
    """
    A factory class to load and run keys lookup services for different
    models/suppliers.
    """
    @classmethod
    def get_model_info(cls, model_version_file_path):
        """
        Get model information from the model version file.
        """
        with io.open(model_version_file_path, 'r', encoding='utf-8') as f:
            return next(csv.DictReader(
                f, fieldnames=['supplier_id', 'model_id', 'model_version']
            ))

    @classmethod
    def get_lookup_package(cls, lookup_package_path):
        """
        Returns the lookup service parent package (called `keys_server` and
        located in `src` in the model keys server Git repository or in
        `var/www/oasis` in the keys server Docker container) from the given
        path.
        """
        if lookup_package_path and not os.path.isabs(lookup_package_path):
            lookup_package_path = os.path.abspath(lookup_package_path)

        parent_dir = os.path.abspath(os.path.dirname(lookup_package_path))
        package_name = re.sub(r'\.py$', '', os.path.basename(lookup_package_path))

        sys.path.insert(0, parent_dir)
        lookup_package = importlib.import_module(package_name)
        imp.reload(lookup_package)
        sys.path.pop(0)

        return lookup_package

    @classmethod
    def get_lookup_class_instance(cls, lookup_package, keys_data_path, model_info):
        """
        Get the keys lookup class instance.
        """
        klc = getattr(lookup_package, '{}KeysLookup'.format(model_info['model_id']))

        return klc(
            keys_data_directory=keys_data_path,
            supplier=model_info['supplier_id'],
            model_name=model_info['model_id'],
            model_version=model_info['model_version']
        )

    @classmethod
    def get_model_exposures(cls, model_exposures=None, model_exposures_file_path=None):
        """
        Get the model exposures/location file data as a pandas dataframe given
        either the path of the model exposures file or the string contents of
        such a file.
        """
        if model_exposures_file_path:
            loc_df = pd.read_csv(os.path.abspath(model_exposures_file_path), float_precision='high')
        elif model_exposures:
            loc_df = pd.read_csv(six.StringIO(model_exposures), float_precision='high')
        else:
            raise OasisException('Either model_exposures_file_path or model_exposures must be specified')

        loc_df = loc_df.where(loc_df.notnull(), None)
        loc_df.columns = loc_df.columns.str.lower()

        return loc_df

    @classmethod
    def write_oasis_keys_file(cls, records, output_file_path, id_col='id'):
        """
        Writes an Oasis keys file from an iterable of keys records.
        """
        heading_row = OrderedDict([
            (id_col, 'LocID'),
            ('peril_id', 'PerilID'),
            ('coverage_type', 'CoverageTypeID'),
            ('area_peril_id', 'AreaPerilID'),
            ('vulnerability_id', 'VulnerabilityID'),
        ])

        pd.DataFrame(
            columns=heading_row.keys(),
            data=[heading_row] + records,
        ).to_csv(
            output_file_path,
            index=False,
            encoding='utf-8',
            header=False,
        )

        return output_file_path, len(records)

    @classmethod
    def write_oasis_keys_errors_file(cls, records, output_file_path, id_col='id'):
        """
        Writes an Oasis keys errors file from an iterable of keys records.
        """
        heading_row = OrderedDict([
            (id_col, 'LocID'),
            ('peril_id', 'PerilID'),
            ('coverage_type', 'CoverageTypeID'),
            ('message', 'Message'),
        ])

        pd.DataFrame(
            columns=heading_row.keys(),
            data=[heading_row] + records,
        ).to_csv(
            output_file_path,
            index=False,
            encoding='utf-8',
            header=False,
        )

        return output_file_path, len(records)

    @classmethod
    def write_json_keys_file(cls, records, output_file_path):
        """
        Writes the keys records as a simple list to file.
        """
        with io.open(output_file_path, 'w', encoding='utf-8') as f:
            f.write(u'{}'.format(json.dumps(records, sort_keys=True, indent=4, ensure_ascii=False)))

            return output_file_path, len(records)

    @classmethod
    def create(
        cls,
        model_keys_data_path=None,
        model_version_file_path=None,
        lookup_package_path=None,
        lookup_config=None,
        lookup_config_json=None,
        lookup_config_fp=None,
        lookup_type='combined',
        loc_id_col='id'
    ):
        """
        Creates a keys lookup class instance for the given model and supplier -
        local file paths are required for the model keys data folder, the model
        version file and the Git repository for the model keys server. Returns a
        pair ``(model_info, klc)``, where ``model_info`` is a dictionary holding
        model information from the model version file and `klc` is the lookup
        service class instance for the model.
        """
        if (lookup_config or lookup_config_json or lookup_config_fp):
            lookup = OasisLookup(
                config=lookup_config,
                config_json=lookup_config_json,
                config_fp=lookup_config_fp,
                loc_id_col=loc_id_col
            )
            model_info = lookup.config.get('model')
            if lookup_type == 'base':
                lookup = OasisBaseLookup(
                    config=lookup_config,
                    config_json=lookup_config_json,
                    config_fp=lookup_config_fp
                )
                return lookup.config.get('model'), lookup
            elif lookup_type == 'combined':
                return model_info, lookup
            elif lookup_type == 'peril':
                return model_info, lookup.peril_lookup
            elif lookup_type == 'vulnerability':
                return model_info, lookup.vulnerability_lookup
        else:
            _model_keys_data_path = as_path(model_keys_data_path, 'model_keys_data_path', preexists=True)
            _model_version_file_path = as_path(model_version_file_path, 'model_version_file_path', preexists=True)
            _lookup_package_path = as_path(lookup_package_path, 'lookup_package_path', preexists=True)

            model_info = cls.get_model_info(_model_version_file_path)
            lookup_package = cls.get_lookup_package(_lookup_package_path)
        
            return model_info, cls.get_lookup_class_instance(lookup_package, _model_keys_data_path, model_info)

    @classmethod
    def get_keys(
        cls,
        lookup=None,
        model_exposures=None,
        model_exposures_file_path=None,
        success_only=True
    ):
        """
        Generates keys records (JSON) for the given model and supplier -
        requires an instance of the lookup service (which can be created using
        the `create` method in this factory class), and either the model
        location file path or the string contents of such a file.

        The optional keyword argument ``success_only`` indicates whether only
        records with successful lookups should be returned (default), or all
        records.
        """
        if not (model_exposures or model_exposures_file_path):
            raise OasisException('No model exposures provided')

        model_loc_df = cls.get_model_exposures(
            model_exposures_file_path=model_exposures_file_path,
            model_exposures=model_exposures
        )

        for record in lookup.process_locations(model_loc_df):
            if success_only:
                if record['status'].lower() == KEYS_STATUS_SUCCESS:
                    yield record
            else:
                yield record

    @classmethod
    def get_results(
        cls,
        lookup,
        model_exposures=None,
        model_exposures_fp=None,
        successes_only=False,
        **kwargs
    ):
        """
        Generates lookup results (dicts) for the given model and supplier -
        requires a lookup instance (which can be created using the `create2`
        method in this factory class), and the model exposures/locations
        dataframe.

        The optional keyword argument ``success_only`` indicates whether only
        results with successful lookup status should be returned (default),
        or all results.
        """
        if not (model_exposures or model_exposures_fp):
            raise OasisException('No model exposures data or file path provided')

        peril_config = lookup.config.get('peril')
        if not peril_config:
            raise OasisException('No peril config defined in the lookup config')

        _model_exposures_fp = as_path(model_exposures_fp, 'model_exposures_fp', preexists=False)

        loc_config = lookup.config.get('locations') or {}
        src_type = 'csv'

        kwargs = {
            'src_data': model_exposures,
            'src_fp': _model_exposures_fp,
            'src_type': 'csv',
            'non_na_cols': tuple(loc_config.get('non_na_cols') or ()),
            'col_dtypes': loc_config.get('col_dtypes') or {},
            'sort_col': loc_config.get('sort_col'),
            'sort_ascending': loc_config.get('sort_ascending')
        }

        model_exposures_df =  get_dataframe(**kwargs)

        locations = (loc for _, loc in model_exposures_df.iterrows())

        for result in lookup.bulk_lookup(locations):
            if successes_only:
                if result['status'].lower() == KEYS_STATUS_SUCCESS:
                    yield result
            else:
                yield result

    @classmethod
    def save_keys(
        cls,
        lookup=None,
        keys_id_col='id',
        keys_file_path=None,
        keys_errors_file_path=None,
        keys_format='oasis',
        model_exposures=None,
        model_exposures_file_path=None,
    ):
        """
        Writes a keys file, and optionally a keys error file, for the keys
        generated by the lookup service for the given model, supplier and
        exposure sfile - requires a lookup service instance (which can be
        created using the `create` method in this factory class), the path of
        the model location file, the path of the keys file, and the format of
        the output file which can be an Oasis keys file (``oasis``) or a
        simple listing of the records to file (``json``).

        The optional keyword argument ``keys_error_file_path`` if present
        indicates that all keys records, whether for locations with successful
        or unsuccessful lookups, will be generated and written to separate
        files. A keys record with a successful lookup will have a `status`
        field whose value will be `success`, otherwise the record will have
        a `status` field value of `failure` or `nomatch`.

        If ``keys_errors_file_path`` is not present then the method returns a
        pair ``(p, n)`` where ``p`` is the keys file path and ``n`` is the
        number of "successful" keys records written to the keys file, otherwise
        it returns a quadruple ``(p1, n1, p2, n2)`` where ``p1`` is the keys
        file path, ``n1`` is the number of "successful" keys records written to
        the keys file, ``p2`` is the keys errors file path and ``n2`` is the
        number of "unsuccessful" keys records written to keys errors file.
        """
        if not (model_exposures or model_exposures_file_path):
            raise OasisException('No model exposures or model exposures file path provided')

        _keys_file_path = as_path(keys_file_path, 'keys_file_path', preexists=False)
        _keys_errors_file_path = as_path(keys_errors_file_path, 'keys_errors_file_path', preexists=False)
        _model_exposures_file_path = as_path(model_exposures_file_path, 'model_exposures_file_path', preexists=False)

        keys = cls.get_keys(
            lookup=lookup,
            model_exposures=model_exposures,
            model_exposures_file_path=_model_exposures_file_path,
            success_only=(True if not keys_errors_file_path else False)
        )

        successes = []
        nonsuccesses = []
        for k in keys:
            successes.append(k) if k['status'] == KEYS_STATUS_SUCCESS else nonsuccesses.append(k)

        if keys_format == 'json':
            if _keys_error_file_path:
                fp1, n1 = cls.write_json_keys_file(successes, _keys_file_path)
                fp2, n2 = cls.write_json_keys_file(nonsuccesses, _keys_errors_file_path)
                return fp1, n1, fp2, n2
            return cls.write_json_keys_file(successes, _keys_file_path)
        elif keys_format == 'oasis':
            if _keys_errors_file_path:
                fp1, n1 = cls.write_oasis_keys_file(successes, _keys_file_path, id_col=keys_id_col)
                fp2, n2 = cls.write_oasis_keys_errors_file(nonsuccesses, _keys_errors_file_path, id_col=keys_id_col)
                return fp1, n1, fp2, n2
            return cls.write_oasis_keys_file(successes, _keys_file_path, id_col=keys_id_col)
        else:
            raise OasisException("Unrecognised keys file output format - valid formats are 'oasis' or 'json'")

    @classmethod
    def save_results(
        cls,
        lookup,
        successes_fp,
        errors_fp=None,
        model_exposures=None,
        model_exposures_fp=None,
        format='oasis'
    ):
        """
        Writes a keys file, and optionally a keys error file, for the keys
        generated by the lookup service for the given model, supplier and
        exposure sfile - requires a lookup service instance (which can be
        created using the `create` method in this factory class), the path of
        the model location file, the path of the keys file, and the format of
        the output file which can be an Oasis keys file (``oasis``) or a
        simple listing of the records to file (``json``).

        The optional keyword argument ``keys_error_file_path`` if present
        indicates that all keys records, whether for locations with successful
        or unsuccessful lookups, will be generated and written to separate
        files. A keys record with a successful lookup will have a `status`
        field whose value will be `success`, otherwise the record will have
        a `status` field value of `failure` or `nomatch`.

        If ``keys_errors_file_path`` is not present then the method returns a
        pair ``(p, n)`` where ``p`` is the keys file path and ``n`` is the
        number of "successful" keys records written to the keys file, otherwise
        it returns a quadruple ``(p1, n1, p2, n2)`` where ``p1`` is the keys
        file path, ``n1`` is the number of "successful" keys records written to
        the keys file, ``p2`` is the keys errors file path and ``n2`` is the
        number of "unsuccessful" keys records written to keys errors file.
        """
        if not (model_exposures or model_exposures_fp):
            raise OasisException('No model exposures data or file path provided')

        mfp = as_path(model_exposures_fp, 'model_exposures_fp', preexists=False)

        sfp = as_path(successes_fp, 'successes_fp', preexists=False)
        efp = as_path(errors_fp, 'errors_fp', preexists=False)

        results = None

        try:
            config = lookup.config
        except AttributeError:
            results = cls.get_keys(
                lookup=lookup,
                model_exposures=model_exposures,
                model_exposures_file_path=mfp,
                success_only=(False if efp else True)
            )
        else:
            results = cls.get_results(
                lookup,
                model_exposures=model_exposures,
                model_exposures_fp=mfp,
                successes_only=(False if efp else True)
            )

        successes = []
        nonsuccesses = []
        for r in results:
            successes.append(r) if r['status'] == KEYS_STATUS_SUCCESS else nonsuccesses.append(r)

        if format == 'json':
            if efp:
                fp1, n1 = cls.write_json_keys_file(successes, sfp)
                fp2, n2 = cls.write_json_keys_file(nonsuccesses, efp)
                return fp1, n1, fp2, n2
            return cls.write_json_keys_file(successes, sfp)
        elif format == 'oasis':
            loc_id_col = None
            try:
                loc_id_col = lookup.loc_id_col
            except AttributeError:
                loc_id_col = 'id'
            else:
                loc_id_col = loc_id_col.lower()
            if efp:
                fp1, n1 = cls.write_oasis_keys_file(successes, sfp, id_col=loc_id_col)
                fp2, n2 = cls.write_oasis_keys_errors_file(nonsuccesses, efp, id_col=loc_id_col)
                return fp1, n1, fp2, n2
            return cls.write_oasis_keys_file(successes, sfp, id_col=loc_id_col)
        else:
            raise OasisException("Unrecognised lookup file output format - valid formats are 'oasis' or 'json'")


class OasisLookup(OasisBaseLookup):
    """
    Combined peril and vulnerability lookup
    """
    @oasis_log()
    def __init__(
        self,
        config=None,
        config_json=None,
        config_fp=None,
        areas=None,
        peril_areas=None,
        peril_areas_index=None,
        peril_areas_index_props=None,
        loc_to_global_areas_boundary_min_distance=0,
        vulnerabilities=None,
        loc_id_col='id'
    ):
        super(self.__class__, self).__init__(
            config=config,
            config_json=config_json,
            config_fp=config_fp
        )

        loc_config = self.config.get('locations')
        self.loc_id_col = str.lower(str(loc_config.get('id_col') or loc_id_col))

        self.peril_lookup = OasisPerilLookup(
            config=self.config,
            areas=areas,
            peril_areas=peril_areas,
            peril_areas_index=peril_areas_index,
            peril_areas_index_props=peril_areas_index_props,
            loc_to_global_areas_boundary_min_distance=loc_to_global_areas_boundary_min_distance,
            loc_id_col=self.loc_id_col
        )

        self.peril_area_id_key = str(str(self.config['peril'].get('peril_area_id_col') or '') or 'peril_area_id').lower()

        self.vulnerability_id_key = str(str(self.config['vulnerability'].get('vulnerability_id_col')) or 'vulnerability_id').lower()

        self.vulnerability_lookup = OasisVulnerabilityLookup(
            config=self.config,
            vulnerabilities=vulnerabilities,
            loc_id_col=self.loc_id_col
        )

    def lookup(self, loc, peril_id, coverage_type):

        loc_id_col = self.loc_id_col
        loc_id = loc.get(loc_id_col) or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

        plookup = self.peril_lookup.lookup(loc, peril_id, coverage_type)

        past = plookup['status']
        pamsg = plookup['message']
        paid = plookup['peril_area_id']
        
        vlookup = self.vulnerability_lookup.lookup(loc, peril_id, coverage_type)

        vlnst = vlookup['status']
        vlnmsg = vlookup['message']
        vlnid = vlookup['vulnerability_id']

        vlookup.pop('status')
        vlookup.pop('message')
        vlookup.pop('vulnerability_id')

        # Could optionally call the status lookup method, but faster
        # to avoid or minimise outside function calls in a `for` loop
        status = (
            KEYS_STATUS_SUCCESS if past == vlnst == KEYS_STATUS_SUCCESS
            else (KEYS_STATUS_FAIL if (past == KEYS_STATUS_FAIL or vlnst == KEYS_STATUS_FAIL) else KEYS_STATUS_NOMATCH)
        )
        
        message = '{}; {}'.format(pamsg, vlnmsg)

        return {
            k:v for k, v in itertools.chain(
                (
                    (loc_id_col, loc_id),
                    ('peril_id', peril_id),
                    ('coverage_type', coverage_type),
                    (self.peril_area_id_key, paid),
                    (self.vulnerability_id_key, vlnid),
                    ('status', status),
                    ('message', message),
                ),
                six.iteritems(vlookup)
            )
        }


class OasisPerilLookup(OasisBaseLookup):
    """
    Single peril, single coverage type, lon/lat point-area poly lookup using
    an Rtree index to store peril areas - index entries are
    #
    #    (peril area ID, peril area bounds)
    #
    # pairs. Areas must be represented as polygons with vertices which are
    # lon/lat coordinates, and be passed in to the constructor as a list,
    # tuple, or generator of triples of the form
    #
    #    (peril area ID, polygon lon/lat vertices, dict with optional properties)
    #
    # An optional distance measure ``loc_to_global_areas_boundary_min_distance`` can be passed
    # that defines how far, in an abstract unit, a given lon/lat location can be
    # from the boundary of the polygon containing all the individual peril area
    # polygons in order to be assigned an peril area ID. By default this distance
    # is 0, which means any lon/lat location outside the polygon containing all
    # peril area polygons will not be assigned a peril area ID.
    """

    @oasis_log()
    def __init__(
        self,
        areas=None,
        config=None,
        config_json=None,
        config_fp=None,
        loc_to_global_areas_boundary_min_distance=0,
        peril_areas=None,
        peril_areas_index=None,
        peril_areas_index_fp=None,
        peril_areas_index_props=None,
        loc_id_col='id'
    ):
        super(self.__class__, self).__init__(config=config, config_json=config_json, config_fp=config_fp)

        peril_config = self.config.get('peril') or {}

        if areas or peril_areas or peril_config:
            if peril_areas_index:
                self.peril_areas_index = peril_areas_index
                self.peril_areas_index_props = self.peril_areas_index_props.properties.as_dict()
            elif (areas or peril_areas):
                self.index_props = (
                    peril_areas_index_props or 
                    peril_config.get('rtree_index') or
                    DEFAULT_RTREE_INDEX_PROPS
                )
                self.peril_areas_index = PerilAreasIndex(areas=areas, peril_areas=peril_areas, properties=self.index_props)
            else:
                areas_rtree_index_config = peril_config.get('rtree_index') or {}
                index_fp = as_path(peril_areas_index_fp or areas_rtree_index_config.get('filename'), 'index_fp', preexists=False)
                if index_fp:
                    idx_ext = areas_rtree_index_config.get('idx_extension') or 'idx'
                    dat_ext = areas_rtree_index_config.get('dat_extension') or 'dat'
                    if not (os.path.exists('{}.{}'.format(index_fp, idx_ext)) or os.path.exists('{}.{}'.format(index_fp, dat_ext))):
                        raise OasisException('No Rtree file index {}.{{idx_ext, dat_ext}} found'.format(index_fp))
                    self.peril_areas_index = PerilAreasIndex(fp=index_fp)
                    self.peril_areas_index_props = self.peril_areas_index.properties.as_dict()

            self.peril_areas_boundary = box(*self.peril_areas_index.bounds, ccw=False)

            _centroid = self.peril_areas_boundary.centroid
            self.peril_areas_centre = _centroid.x, _centroid.y

            self.loc_to_global_areas_boundary_min_distance = (
                loc_to_global_areas_boundary_min_distance or 
                self.config['peril'].get('loc_to_global_areas_boundary_min_distance') or 0
            )

        if self.config.get('locations'):
            self.loc_id_col = str.lower(str(self.config['locations'].get('id_col') or loc_id_col))
            self.loc_coords_x_col = str.lower(str(self.config['locations'].get('coords_x_col')) or 'lon')
            self.loc_coords_y_col = str.lower(str(self.config['locations'].get('coords_y_col')) or 'lat')
            self.loc_coords_x_bounds = tuple(self.config['locations'].get('coords_x_bounds') or ()) or (-180, 180)
            self.loc_coords_y_bounds = tuple(self.config['locations'].get('coords_y_bounds') or ()) or (-90, 90)

    def lookup(self, loc, peril_id, coverage_type):
        """
        Area peril lookup for an individual lon/lat location item, which can be
        provided as a dict or a Pandas series. The data structure should contain
        the keys `lon` or `longitude` for longitude and `lat` or `latitude` for
        latitude.
        """
        idx = self.peril_areas_index
        boundary = self.peril_areas_boundary
        loc_to_areas_min_dist = self.loc_to_global_areas_boundary_min_distance

        loc_id_col = self.loc_id_col

        loc_id = loc.get(loc_id_col) or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

        loc_x_col = self.loc_coords_x_col
        loc_y_col = self.loc_coords_y_col
        loc_x_bounds = self.loc_coords_x_bounds
        loc_y_bounds = self.loc_coords_y_bounds

        x = loc.get(loc_x_col)
        y = loc.get(loc_y_col)

        _lookup = lambda loc_id, x, y, st, perid, covtype, paid, pabnds, pacoords, msg: {
            loc_id_col: loc_id,
            loc_x_col: x,
            loc_y_col: y,
            'peril_id': perid,
            'coverage_type': covtype,
            'status': st,
            'peril_area_id': paid,
            'area_peril_id': paid,
            'area_bounds': pabnds,
            'area_coordinates': pacoords,
            'message': msg
        }

        try:
            x = float(x)
            y = float(y)
            if not ((loc_x_bounds[0] <= x <= loc_x_bounds[1]) and (loc_y_bounds[0] <= y <= loc_y_bounds[1])):
                raise ValueError('{}/{} out of bounds'.format(loc_x_col, loc_y_col))
        except (ValueError, TypeError) as e:
            msg = (
                'Peril area lookup: invalid {}/{} ({}, {}) - {}'
                .format(loc_x_col, loc_y_col, x, y, str(e))
            )
            return _lookup(loc_id, x, y, KEYS_STATUS_FAIL, peril_id, coverage_type, None, None, None, msg)

        st = KEYS_STATUS_NOMATCH
        msg = 'No peril area match'
        paid = None
        pabnds = None
        pacoords = None
        point = x, y

        try:
            results = list(idx.intersection(point, objects='raw'))

            if not results:
                raise IndexError

            for _perid, _covtype, _paid, _pabnds, _pacoords in results:
                if (peril_id, coverage_type) == (_perid, _covtype):
                    paid, pabnds, pacoords = _paid, _pabnds, _pacoords
                    break

            if paid == None:
                raise IndexError
        except IndexError:
            try:
                results = list(idx.nearest(point, objects='raw'))

                if not results:
                    raise IndexError

                for _perid, _covtype, _paid, _pabnds, _pacoords in results:
                    if (peril_id, coverage_type) == (_perid, _covtype):
                        paid, pabnds, pacoords = _paid, _pabnds, _pacoords
                        break

                if paid == None:
                    msg = 'No intersecting or nearest peril area found for peril ID {} and coverage type {}'.format(peril_id, coverage_type)
                    return _lookup(loc_id, x, y, KEYS_STATUS_NOMATCH, peril_id, coverage_type, None, None, None, msg)
            except IndexError:
                pass
            else:
                p = Point(x, y)
                min_dist = p.distance(boundary)
                if min_dist > loc_to_areas_min_dist:
                    msg = (
                        'Peril area lookup: location is {} units from the '
                        'peril areas global boundary -  the required minimum '
                        'distance is {} units'
                        .format(min_dist, loc_to_areas_min_dist)
                    )
                    return _lookup(loc_id, x, y, KEYS_STATUS_FAIL, peril_id, coverage_type, None, None, None, msg)
                st = KEYS_STATUS_SUCCESS
                msg = (
                    'Successful peril area lookup: {}'.format(paid)
                )
        except RTreeError as e:
            return _lookup(loc_id, x, y, KEYS_STATUS_FAIL, peril_id, coverage_type, None, None, None, str(e))
        else:
            st = KEYS_STATUS_SUCCESS
            msg = 'Successful peril area lookup: {}'.format(paid)

        return _lookup(loc_id, x, y, st, peril_id, coverage_type, paid, pabnds, pacoords, msg)


class OasisVulnerabilityLookup(OasisBaseLookup):
    """
    Simple key-value based vulnerability lookup
    """

    @oasis_log()
    def __init__(
        self,
        config=None,
        config_json=None,
        config_fp=None,
        vulnerabilities=None,
        loc_id_col='id'
    ):
        super(self.__class__, self).__init__(config=config, config_json=config_json, config_fp=config_fp)

        if vulnerabilities or self.config.get('vulnerability'):
            self.col_dtypes, self.key_cols, self.vuln_id_col, self.vulnerabilities = self.get_vulnerabilities(vulnerabilities=vulnerabilities)

        if self.config.get('locations'):
            self.loc_id_col = str.lower(str(self.config['locations'].get('id_col') or loc_id_col))

    @oasis_log()
    def get_vulnerabilities(self, vulnerabilities=None):
        if not self.config:
            raise OasisException(
                'No lookup configuration provided or set - use `get_config` '
                'on this instance to set it and provide either an actual '
                'model config dict (use `model_config` argument), or a model '
                'config JSON string (use `model_config_json` argument, or a '
                'model config JSON file path (use `model_config_fp` argument)'
            )

        vuln_config = self.config.get('vulnerability')

        if not vuln_config:
            raise OasisException('No vulnerability config set in the lookup config')

        col_dtypes = vuln_config.get('col_dtypes')

        if not col_dtypes:
            raise OasisException(
                'Vulnerability file column data types must be defined as a '
                '(col, data type) dict in the vulnerability section of the '
                'lookup config'
            )

        col_dtypes = {
            k.lower():getattr(builtins, v) for k, v in six.iteritems(col_dtypes)
        }

        key_cols = vuln_config.get('key_cols')

        if not vuln_config.get('key_cols'):
            raise OasisException(
                'The vulnerability file key column names must be listed in the '
                'vulnerability section of the lookup config'
            )

        key_cols = tuple(col.lower() for col in key_cols)

        vuln_id_col = str(str(self.config['vulnerability'].get('vulnerability_id_col')) or 'vulnerability_id').lower()

        def _vuln_dict(vulns_seq, key_cols, vuln_id_col):
            return (
                {v[key_cols[0]]:(v.get(vuln_id_col) or v.get('vulnerability_id')) for _, v in vulns_seq} if len(key_cols) == 1
                else OrderedDict(
                    {tuple(v[key_cols[i]] for i in range(len(key_cols))):(v.get(vuln_id_col) or v.get('vulnerability_id')) for v in vulns_seq}
                )
            )

        if vulnerabilities:
            return col_dtypes, key_cols, vuln_id_col, _vuln_dict(enumerate(vulnerabilities), key_cols)

        src_fp = vuln_config.get('file_path')

        if not src_fp:
            raise OasisException(
                'No vulnerabilities file path provided in the lookup config'
            )

        if not os.path.isabs(src_fp):
            src_fp = os.path.abspath(src_fp)

        self.config['vulnerability']['file_path'] = src_fp

        src_type = str(str(vuln_config.get('file_type')) or 'csv').lower()

        float_precision = 'high' if vuln_config.get('float_precision_high') else None

        non_na_cols = vuln_config.get('non_na_cols') or tuple(col.lower() for col in list(key_cols) + [vuln_id_col])

        sort_col = vuln_config.get('sort_col') or vuln_id_col
        sort_ascending = vuln_config.get('sort_ascending')

        vuln_df = get_dataframe(
            src_fp=src_fp,
            src_type=src_type,
            float_precision=float_precision,
            lowercase_cols=True,
            index_col=True,
            non_na_cols=non_na_cols,
            col_dtypes=col_dtypes,
            sort_col=sort_col,
            sort_ascending=sort_ascending
        )

        return col_dtypes, key_cols, vuln_id_col, _vuln_dict((v for _, v in vuln_df.iterrows()), key_cols, vuln_id_col)

    def lookup(self, loc, peril_id, coverage_type):
        """
        Vulnerability lookup for an individual location item, which could be a dict or a
        Pandas series.
        """
        loc_id_col = self.loc_id_col
        loc_id = loc.get(loc_id_col) or int(uuid.UUID(bytes=os.urandom(16)).hex[:16], 16)

        key_cols = self.key_cols
        col_dtypes = self.col_dtypes
        vuln_id_col = self.vuln_id_col

        loc_key_col_values = OrderedDict({
            key_col: loc.get(key_col) for key_col in key_cols
        })

        if not loc_key_col_values['peril_id']:
            loc_key_col_values['peril_id'] = peril_id

        if not loc_key_col_values['coverage_type']:
            loc_key_col_values['coverage_type'] = loc.get('coverage') or coverage_type

        _lookup = lambda loc_id, vlnperid, vlncovtype, vlnst, vlnid, vlnmsg: {
            k:v for k, v in itertools.chain(
                (
                    (loc_id_col, loc_id),
                    ('peril_id', vlnperid),
                    ('coverage_type', vlncovtype),
                    ('status', vlnst),
                    ('vulnerability_id', vlnid),
                    ('message', vlnmsg)
                ),
                six.iteritems(loc_key_col_values)
            )
        }

        try:
            for key_col in key_cols:
                key_col_dtype = col_dtypes[key_col]
                key_col_dtype(loc_key_col_values[key_col])
        except (TypeError, ValueError):
            return _lookup(loc_id, peril_id, coverage_type, KEYS_STATUS_FAIL, None, 'Vulnerability lookup: invalid key column value(s) for location')

        vlnperid = peril_id
        vlncovtype = coverage_type
        vlnst = KEYS_STATUS_NOMATCH
        vlnmsg = 'No vulnerability match'
        vlnid = None

        try:
            vlnid = (
                self.vulnerabilities[tuple(loc_key_col_values[col] for col in key_cols)] if len(key_cols) > 1
                else self.vulnerabilities[loc[key_cols[0]]]
            )
        except KeyError:
            pass
        else:
            vlnperid = peril_id
            vlncovtype = coverage_type
            vlnst = KEYS_STATUS_SUCCESS
            vlnmsg = 'Successful vulnerability lookup: {}'.format(vlnid)

        return _lookup(loc_id, vlnperid, vlncovtype, vlnst, vlnid, vlnmsg)



import socket
from unittest import TestCase

from tempfile import NamedTemporaryFile

from mock import patch, Mock

from oasislmf.utils.conf import load_ini_file, replace_in_file
from oasislmf.utils.exceptions import OasisException


class LoadInIFile(TestCase):
    def test_values_are_bool___values_are_correctly_converted_to_bool_value(self):
        with NamedTemporaryFile(mode='w') as f:
            f.writelines([
                '[section]\n',
                'a = True\n',
                'b = False\n',
            ])
            f.flush()

            conf = load_ini_file(f.name)

            self.assertTrue(conf['a'])
            self.assertFalse(conf['b'])

    def test_values_are_int___values_are_correctly_converted_to_int_value(self):
        with NamedTemporaryFile(mode='w') as f:
            f.writelines([
                '[section]\n',
                'a = 1\n',
                'b = 2\n',
            ])
            f.flush()

            conf = load_ini_file(f.name)

            self.assertEqual(1, conf['a'])
            self.assertEqual(2, conf['b'])

    def test_values_are_float___value_are_correctly_converted_to_int_value(self):
        with NamedTemporaryFile(mode='w') as f:
            f.writelines([
                '[section]\n',
                'a = 1.1\n',
                'b = 2.2\n',
            ])
            f.flush()

            conf = load_ini_file(f.name)

            self.assertEqual(1.1, conf['a'])
            self.assertEqual(2.2, conf['b'])

    def test_values_are_ip_addresses___values_are_converted_into_ip_string_format(self):
        with NamedTemporaryFile(mode='w') as f:
            f.writelines([
                '[section]\n',
                'a = 127.0.0.1\n',
                'b = 127.127.127.127\n',
            ])
            f.flush()

            conf = load_ini_file(f.name)

            ipf = lambda s: socket.inet_ntoa(socket.inet_aton(s))

            self.assertEqual(ipf('127.0.0.1'), conf['a'])
            self.assertEqual(ipf('127.127.127.127'), conf['b'])

    def test_values_are_string_values___values_are_unchanged(self):
        with NamedTemporaryFile(mode='w') as f:
            f.writelines([
                '[section]\n',
                'a = first.value\n',
                'b = another value\n',
            ])
            f.flush()

            conf = load_ini_file(f.name)

            self.assertEqual('first.value', conf['a'])
            self.assertEqual('another value', conf['b'])

    def test_io_error_is_raised_when_opening_file___exception_is_converted_to_oasis_exception(self):
        def raising_function(*args, **kwargs):
            raise IOError()

        with patch('io.open', Mock(side_effect=raising_function)), self.assertRaises(OasisException):
            load_ini_file('file_name')


class ReplaceInFile(TestCase):
    def test_more_var_names_are_given_than_values___error_is_raised(self):
        with self.assertRaises(OasisException):
            replace_in_file('first_path', 'second_path', ['fist_arg', 'second_arg'], ['first_val'])

    def test_more_var_values_are_given_than_values___error_is_raised(self):
        with self.assertRaises(OasisException):
            replace_in_file('first_path', 'second_path', ['fist_arg'], ['first_val', 'second_val'])

    def test_input_file_does_not_include_any_var_names___file_is_unchanged(self):
        with NamedTemporaryFile(mode='w') as input_file, NamedTemporaryFile(mode='r') as output_file:
            input_file.writelines([
                'some_var some_val\n',
            ])
            input_file.flush()

            replace_in_file(input_file.name, output_file.name, ['first_arg', 'second_arg'], ['first_val', 'second_val'])

            output_file.seek(0)
            data = output_file.read()

            self.assertEqual('some_var some_val\n', data)

    def test_input_file_includes_some_var_names___input_names_are_replaced_with_values(self):
        with NamedTemporaryFile(mode='w') as input_file, NamedTemporaryFile(mode='r') as output_file:
            input_file.writelines([
                'some_var first_arg\n',
            ])
            input_file.flush()

            replace_in_file(input_file.name, output_file.name, ['first_arg', 'second_arg'], ['first_val', 'second_val'])

            output_file.seek(0)
            data = output_file.read()

            self.assertEqual('some_var first_val\n', data)

    def test_io_error_is_raised_when_opening_file___exception_is_converted_to_oasis_exception(self):
        def raising_function(*args, **kwargs):
            raise IOError()

        with patch('io.open', Mock(side_effect=raising_function)), self.assertRaises(OasisException):
            replace_in_file('in', 'out', ['first_arg', 'second_arg'], ['first_val', 'second_val'])

    def test_os_error_is_raised_when_opening_file___exception_is_converted_to_oasis_exception(self):
        def raising_function(*args, **kwargs):
            raise OSError()

        with patch('io.open', Mock(side_effect=raising_function)), self.assertRaises(OasisException):
            replace_in_file('in', 'out', ['first_arg', 'second_arg'], ['first_val', 'second_val'])



from tempfile import NamedTemporaryFile
from unittest import TestCase

from oasislmf.utils.csv import read_csv


class GetCsvRowsAsDicts(TestCase):
    def test_file_has_no_rows___result_is_empty(self):
        with NamedTemporaryFile('w') as f:
            f.writelines([
                'first,second\n',
            ])
            f.flush()

            self.assertEqual(0, len(list(read_csv(f.name))))

    def test_no_meta_is_given___base_csv_data_is_given(self):
        with NamedTemporaryFile('w') as f:
            f.writelines([
                'first,second\n',
                'a,b\n',
                'c,d\n',
            ])
            f.flush()

            data = list(read_csv(f.name))

            self.assertEqual(
                data, [
                    {
                        'first': 'a',
                        'second': 'b',
                    },
                    {
                        'first': 'c',
                        'second': 'd',
                    }
                ]
            )

    def test_no_meta_is_given___data_is_transformed_by_the_meta_functions(self):
        with NamedTemporaryFile('w') as f:
            f.writelines([
                'first,second\n',
                'a,1\n',
                'c,2\n',
            ])
            f.flush()

            data = list(read_csv(f.name, csv_meta={
                'FIRST': {'validator': str.upper, 'csv_header': 'first'},
                'SECOND': {'validator': int, 'csv_header': 'second'},
            }))

            self.assertEqual(
                data, [
                    {
                        'FIRST': 'A',
                        'SECOND': 1,
                    },
                    {
                        'FIRST': 'C',
                        'SECOND': 2,
                    }
                ]
            )



import uuid
from unittest import TestCase

import os
import six
from backports.tempfile import TemporaryDirectory
from mock import patch

from oasislmf.cmd import RootCmd
from oasislmf.utils.exceptions import OasisException


def get_command(src_dir=None, dst_dir=None, extras=None):
    kwargs_str = ' '.join('--{} {}'.format(k, v) for k, v in six.iteritems(extras or {}))

    return RootCmd(argv='bin build {} {} {}'.format(kwargs_str, src_dir or '', dst_dir or '').split())


class BuildCmdParseArgs(TestCase):
    def test_source_directory_is_not_supplied___default_is_current_directory(self):
        cmd = get_command()

        args = cmd.parse_args()

        self.assertEqual(args.source, os.path.abspath('.'))

    def test_source_directory_is_supplied___value_is_abs_path_to_the_supplied_direcory(self):
        with TemporaryDirectory() as d:
            cmd = get_command(src_dir=d)

            args = cmd.parse_args()

            self.assertEqual(args.source, os.path.abspath(d))

    def test_source_directory_doesnt_exist___error_is_raised(self):
        cmd = get_command(src_dir='non_existing_dir_{}'.format(uuid.uuid4().hex))

        with self.assertRaises(OasisException):
            cmd.parse_args()

    def test_destination_directory_is_not_supplied___default_is_none(self):
        cmd = get_command()

        args = cmd.parse_args()

        self.assertIsNone(args.destination)

    def test_destination_directory_is_supplied___value_is_abs_path_to_the_supplied_direcory(self):
        with TemporaryDirectory() as src, TemporaryDirectory() as dst:
            cmd = get_command(src_dir=src, dst_dir=dst)

            args = cmd.parse_args()

            self.assertEqual(args.destination, os.path.abspath(dst))

    def test_do_il_is_not_supplied___do_il_is_false(self):
        cmd = get_command()

        args = cmd.parse_args()

        self.assertFalse(args.do_il)

    def test_do_il_is_supplied___do_il_is_true(self):
        cmd = get_command(extras={'do-il': ''})

        args = cmd.parse_args()

        self.assertTrue(args.do_il)

    def test_build_tar_is_not_supplied___build_tar_il_is_false(self):
        cmd = get_command()

        args = cmd.parse_args()

        self.assertFalse(args.build_tar)

    def test_build_tar_is_supplied___build_tar_is_true(self):
        cmd = get_command(extras={'build-tar': ''})

        args = cmd.parse_args()

        self.assertTrue(args.build_tar)


class BuildCmdRun(TestCase):
    @patch('oasislmf.cmd.bin.check_conversion_tools')
    @patch('oasislmf.cmd.bin.check_inputs_directory')
    @patch('oasislmf.cmd.bin.create_binary_files')
    @patch('oasislmf.cmd.bin.create_binary_tar_file')
    def test_dst_dir_is_not_supplied___flow_is_correct_using_src_as_dst(self, create_tar_mock, create_bin_mock, check_ins_mock, check_conv_mock):
        with TemporaryDirectory() as src:
            cmd = get_command(src_dir=src, extras={'build-tar': ''})

            res = cmd.run()

            self.assertEqual(0, res)
            check_conv_mock.assert_called_once_with(do_il=False)
            check_ins_mock.assert_called_once_with(src, do_il=False, check_binaries=False)
            create_bin_mock.assert_called_once_with(src, src, do_il=False)
            create_tar_mock.assert_called_once_with(src)

    @patch('oasislmf.cmd.bin.check_conversion_tools')
    @patch('oasislmf.cmd.bin.check_inputs_directory')
    @patch('oasislmf.cmd.bin.create_binary_files')
    @patch('oasislmf.cmd.bin.create_binary_tar_file')
    def test_dst_dir_is_supplied___flow_is_correct_using_supplied_dir_as_dst(self, create_tar_mock, create_bin_mock, check_ins_mock, check_conv_mock):
        with TemporaryDirectory() as src, TemporaryDirectory() as dst:
            cmd = get_command(src_dir=src, dst_dir=dst, extras={'build-tar': ''})

            res = cmd.run()

            self.assertEqual(0, res)
            check_conv_mock.assert_called_once_with(do_il=False)
            check_ins_mock.assert_called_once_with(src, do_il=False, check_binaries=False)
            create_bin_mock.assert_called_once_with(src, dst, do_il=False)
            create_tar_mock.assert_called_once_with(dst)

    @patch('oasislmf.cmd.bin.check_conversion_tools')
    @patch('oasislmf.cmd.bin.check_inputs_directory')
    @patch('oasislmf.cmd.bin.create_binary_files')
    @patch('oasislmf.cmd.bin.create_binary_tar_file')
    def test_build_tar_is_false___create_tar_file_is_not_called(self, create_tar_mock, create_bin_mock, check_ins_mock, check_conv_mock):
        with TemporaryDirectory() as src:
            cmd = get_command(src_dir=src)

            res = cmd.run()

            self.assertEqual(0, res)
            check_conv_mock.assert_called_once_with(do_il=False)
            check_ins_mock.assert_called_once_with(src, do_il=False, check_binaries=False)
            create_bin_mock.assert_called_once_with(src, src, do_il=False)
            create_tar_mock.assert_not_called()

    @patch('oasislmf.cmd.bin.check_conversion_tools')
    @patch('oasislmf.cmd.bin.check_inputs_directory')
    @patch('oasislmf.cmd.bin.create_binary_files')
    @patch('oasislmf.cmd.bin.create_binary_tar_file')
    def test_do_il_is_true___il_files_are_generated(self, create_tar_mock, create_bin_mock, check_ins_mock, check_conv_mock):
        with TemporaryDirectory() as src:
            cmd = get_command(src_dir=src, extras={'do-il': ''})

            res = cmd.run()

            self.assertEqual(0, res)
            check_conv_mock.assert_called_once_with(do_il=True)
            check_ins_mock.assert_called_once_with(src, do_il=True, check_binaries=False)
            create_bin_mock.assert_called_once_with(src, src, do_il=True)
            create_tar_mock.assert_not_called()



# -*- coding: utf-8 -*-

import zlib

from .exceptions import OasisException

CHUNK_SIZE = 5 * 10 ** 8  # 500 Mb


def compress_data(s):
    """
    Compress large data strings.

    Adapted from a StackOverflow.com solution by Dmitry Skryabin

        https://stackoverflow.com/a/36056646/7556955

    with a modification to set block/chunk size to 500 Mb (5 x 10^8 bytes).
    """

    compressed = b''
    begin = 0
    compressor = zlib.compressobj()

    try:
        while begin < len(s):
            compressed += compressor.compress(s[begin:begin + CHUNK_SIZE])
            begin += CHUNK_SIZE

        compressed += compressor.flush()
    except zlib.error as e:
        raise OasisException(str(e))

    return compressed
